/*
 * KOTETU: the prefetching WWW proxy server
 *
 * proxy_main.c : main body for proxy server.
 *
 *      by k-chinen@is.aist-nara.ac.jp, 1999, 2000
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <pthread.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_POLL_H
#include <poll.h>
#include <sys/poll.h>
#endif
#ifdef HAVE_SELECT_H
#include <sys/select.h>
#endif
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#include <strings.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <assert.h>
#include <errno.h>

/* private headers */

#include "range.h"
#include "msg.h"
#include "mem.h"

#include "socket.h"
#include "sigs.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "misc.h"
#include "http.h"
#include "xsk.h"
#include "sess.h"
#include "squeue.h"
#include "parse.h"
#include "mlist.h"
#include "dhr.h"

#include "reason.h"


#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef DEFAULT_CONFIG_ENVNAME
#define DEFAULT_CONFIG_ENVNAME "KOTETU"
#endif

#ifndef DEFAULT_CONFIGFILE
#define DEFAULT_CONFIGFILE      "/etc/kotetu.conf"
#endif
#ifndef DEFAULT_LOGFILE
#define DEFAULT_LOGFILE         "kotetu.log"
#endif

#ifndef MONITOR_INTERVAL
#define MONITOR_INTERVAL    (600)
#endif

#ifndef MALLOC
#define MALLOC(x)   malloc(x)
#endif
#ifndef FREE
#define FREE(x)     free(x)
#endif

#define KOTETU_VER  "Kotetu-1.0"


/* dummy */

#define cache_show_blockarray(x,y)      ((void)0)
#define cache_list_entry(x)             ((void)0)
int nline;
int nblock;
int linesize;



pthread_mutex_t stat_lock = PTHREAD_MUTEX_INITIALIZER;
time_t boot_date = 0;
time_t service_start_date = 0;
int    session_queue_report_flag = 0;

volatile long long request_sirial_count = 0;
volatile int receive_request_count = 0;
volatile int process_request_count = 0;
volatile int issue_prefetch_count  = 0;
volatile int report_request_count  = 0;
volatile int reject_request_count  = 0;
volatile int ignore_request_count  = 0;

volatile int send_reply_count    = 0;
volatile int hit_reply_count     = 0;
volatile int miss_reply_count    = 0;
volatile int error_reply_count   = 0;

volatile int friend_count       = 0;

volatile int hash_colision      = 0;
volatile int compaction_count   = 0;
volatile int msweeper_count     = 0;
volatile int expire_count       = 0;

volatile int skip_cache_count   = 0;
volatile int alloc_tmp_buf_count = 0;
volatile int alloc_sess_buf_count = 0;

extern volatile int swapload_count;
extern volatile int swapin_count;
extern volatile int swapout_count;
extern volatile int swapload_success;
extern volatile int swapin_success;
extern volatile int swapout_success;

extern volatile int swapout_token;
extern volatile int swapout_token_init;
extern volatile int swapout_token_interval;

extern volatile int dsweeper_seq;
extern volatile int dsweeper_count;
extern volatile int dsweeper_phase;
extern volatile int dsweeper_pos;

extern volatile int checkdir_dir_count;
extern volatile int checkdir_file_read_count;
extern volatile int checkdir_file_handle_count;
extern volatile int checkdir_file_left_count;
extern volatile int checkdir_file_delete_count;
extern volatile int checkdir_file_erase_count;


int read_reply_total   = 0;
int read_reply_success = 0;
int read_reply_fail    = 0;

int cache_newover_count = 0;
int cache_newfree_count = 0;
int cache_newmin_count = 0;



volatile long long intra_incoming = (off_t)0; /* requests from clients */
volatile long long extra_outgoing = (off_t)0; /* requests to super class servers */
volatile long long extra_incoming = (off_t)0; /* replys from super class servers */
volatile long long intra_outgoing = (off_t)0; /* replys to clients */



sim_stat_rec rcode_stat;
long    rcode_stat_count[1000];

sim_stat_rec ret_time;
long    ret_time_count[K(7)];

sim_stat_rec ret_time_short;
long    ret_time_short_count[K(32)];

sim_stat_rec len_perread_stat;
static long server_buffer_stat_count[K(32)];

sim_stat_rec object_size_stat;
static long object_size_stat_count[K(1)];

sim_stat_rec url_first_hash_stat;
static long url_first_hash_stat_count[K(2)];
sim_stat_rec url_second_hash_stat;
static long url_second_hash_stat_count[K(2)];
sim_stat_rec url_third_hash_stat;
static long url_third_hash_stat_count[K(10)];
sim_stat_rec url_fourth_hash_stat;
static long url_fourth_hash_stat_count[K(10)];

sim_stat_rec req_field_type_stat;
static long req_field_type_stat_count[K(10)];

sim_stat_rec object_alloc_stat;
static long object_alloc_stat_count[K(10)];

sim_stat_rec msweep_time;
static long msweep_time_count[K(10)];
sim_stat_rec msweep_time_short;
static long msweep_time_short_count[K(10)];


pthread_t th_listener;
pthread_mutex_t accept_lock = PTHREAD_MUTEX_INITIALIZER;

pthread_t th_handlers[N_HANDLER_MAX];
pthread_t th_dsweepers[N_HANDLER_MAX];
pthread_t th_monitor;

pthread_t th_signalhandler;

sigset_t handler_sig_mask;

int require_sess_length;

int think_queue_length;
sess_queue_t X_think_queue;
sess_queue_t *think_queue;
#if 0
sess_queue_t X_send_queue;
sess_queue_t *send_queue;
sess_queue_t X_fetch_queue;
sess_queue_t *fetch_queue;

sess_queue_t X_prompt_queue;
sess_queue_t *prompt_queue;
#endif

sess_queue_t X_wait_queue;
sess_queue_t *wait_queue;

sess_queue_t X_msweep_queue;
sess_queue_t *msweep_queue;

#ifndef N_SWEEPER_MAX
#define N_SWEEPER_MAX N_HANDLER_MAX
#endif
sess_queue_t X_swapout_queue[N_SWEEPER_MAX];
sess_queue_t *swapout_queue[N_SWEEPER_MAX];

#if 0
sess_queue_t X_swapout_queue;
sess_queue_t *swapout_queue;
#endif
sess_queue_t X_swapin_queue;
sess_queue_t *swapin_queue;


#ifdef CHECK_QUEUE

#define push_sess(q,c) \
    RawTrace("push sess# %d, %p to %p : %d/%d : %s %5d\n",\
        c->sid, c, q, q->count, q->cmax, __FILE__ , __LINE__), sq_push(q,c)
#define pull_sess(q) \
    RawTrace("pull from %p : %s %5d\n", q, __FILE__ , __LINE__), sq_pull(q,c)

#else
#define push_sess(q,c) sq_push(q,c)
#define pull_sess(q) sq_pull(q)
#endif
#define pull_sess_nonblock(q) sq_pull_nonblock(q)



sim_hist_rec comp_hist;
static long comp_hist_area[128];
sim_hist_rec expire_hist;
static long expire_hist_area[128];

extern sim_hist_rec fratio_hist;

/*
 * externals
 */

extern int nhot;
extern int nfriend;

#if 0
#define push_sess(q,c) sq_timedpush(q,c,100000) /* 100 msec */
#endif


int InitSession(int);


/* automatic */
pthread_mutex_t config_lock = PTHREAD_MUTEX_INITIALIZER;

char config_filename[L_PATH];
char *log_filename;
int trace_flag;
int monitor_flag;
int prefetch_flag=1;
int service_portno;
int require_n_handler;
int require_n_dsweeper;
int require_n_msweeper;
int require_chunk_size;
off_t require_total_memory_size;

int require_num_bucket;
int require_level_bucket_skiplist;
int require_size_frag;
int require_ratio_frag_length;

int num_prefetch_referedpages;
int num_prefetch_includedobjs;


int nfrag_obj       = HARD_NFRAG_OBJ;
int nfrag_nwrite    = HARD_NFRAG_NWRITE;
int nfrag_nread     = HARD_NFRAG_NREAD;
int nfrag_dwrite    = HARD_NFRAG_DWRITE;
int nfrag_dread     = HARD_NFRAG_DREAD;

int require_nfrag_obj;
int require_nfrag_nwrite;
int require_nfrag_nread;
int require_nfrag_dwrite;
int require_nfrag_dread;

/* for HR and DHR */
int n_proxy;
int proxy_id;
PROXY_ID *cluster;


char *localdomain_padding;
char *report_hook_hostname;

/* static */
pid_t    wcol_pid;
char     *progname;


extern off_t sess_memory_size;
off_t cache_memory_size;


extern int sq_check_count();
extern int check_emptyentity(bucket_t* bk);


#if 1
/* nanosleep is useful, because it is MT-safe and supports very short time.
 * however, nsec is too short for network I/O.
 * then, msleep is made.
 */
/*
inline
*/
static
int
msleep(int x)
{
    struct timespec w;

    w.tv_sec = 0;
    w.tv_nsec = x * 1000000;   /* convert msec to nsec */

    nanosleep(&w, NULL);

    return 0;
}
#endif



int
check_sq(char *msg)
{
    Trace("# %s: sq_count(free) %d, diff %d, think Q %d/%d, wait Q %d/%d, msweep Q %d/%d\n",
        msg,
        sq_check_count(),
        sq_check_count()
            - (think_queue->count + wait_queue->count + msweep_queue->count),
        think_queue->count, think_queue->cmax,
        wait_queue->count, wait_queue->cmax,
        msweep_queue->count, msweep_queue->cmax);
    return 0;
}

int
CHKTIME(time_t thetime)
{
    int diff;
    diff = time(NULL) - thetime;
    if(diff<0) diff = -diff;
    if(diff>86400) {
        return 0;
    }
    else {
        return 1;
    }
}

#if 0
void
dump_bkinfo()
{
    int i;
    bucket_t *b;

    for(i=0;i<radix_h1;i++) {
        b = &skcache_bucket_base[i];
        Trace(" %3d: entry %4d/%d/%d %08x, frag %4d/%d/%d %08x\n",
            i,
            b->entry_use, b->entry_hwater, b->entry_all, b->entry_freelist,
            b->frag_use, b->frag_hwater, b->frag_all, b->frag_freelist);
    }
}
#endif


/* free list of entry or frag is empty, when the number of use is smaller
 * than high water.
 */
int
check_bkinfo()
{
    int i;
    bucket_t *b;
    int r, c;

    c = 0;
    for(i=0;i<radix_h1;i++) {
        b = &skcache_bucket_base[i];
        r = 0;
        if(b->entry_freelist==NULL && b->entry_use <  b->entry_hwater) {
            r++;
        }
#if 0
        if(b->frag_freelist==NULL && b->frag_use <  b->frag_hwater) {
            r++;
        }
#endif

        if(r) {
#if 0
          Trace(" be %3d: entry %4d/%d/%d %08x, frag %4d/%d/%d %08x\n",
            i,
            b->entry_use, b->entry_hwater, b->entry_all, b->entry_freelist,
            b->frag_use, b->frag_hwater, b->frag_all, b->frag_freelist);
#endif
          Trace(" bk# %d, entry u %4d f %d hw %d all %d, %08x, frag %4d\n",
            i,
            b->entry_use, b->entry_free, b->entry_hwater, b->entry_all,
            b->entry_freelist,
            b->frag_use);

            c++;
        }
    }

    if(c==0)
        return 1;
    else
        return 0;
}

pthread_mutex_t wait_check_lock;

inline int check_timeout(sess_t *sessp, time_t now);

int
check_friend_lock(sess_t *this)
{
    sess_t *cu;
    int n;
    int i;
    int chk;
    int w, d;
    time_t now;
    int idle;

    if(this==NULL || this->cache_entry==NULL) {
        return -1;
    }
Trace("check_friend_lock:\n");
Trace("  target sess# %d, oid# %d\n",
    this->sid, ((skcache_entry_t*)this->cache_entry)->oid);


    MUTEX_LOCK(&wait_check_lock);

    now = time(NULL);

    n = wait_queue->count*2; /* XXX */
    i = 0;
    w = 0;
    d = 0;
    while(i<n) {
        cu = pull_sess_nonblock(wait_queue);
        if(cu==NULL) {
            break;
        }
        idle = now - cu->last_action;

#if 0
  if(cu->cache_entry) {
    Trace("  %3d: sess# %d, %p %s : entry %p, oid# %d, ref %d, owner %p, idle %d\n",
    i,
    cu->sid, cu, MSG_STATE_[cu->state],
    cu->cache_entry,
    ((skcache_entry_t*)cu->cache_entry)->oid,
    ((skcache_entry_t*)cu->cache_entry)->ref,
    ((skcache_entry_t*)cu->cache_entry)->owner, idle
    );
  }
#endif

        if(check_timeout(cu, now)) {
            cu->state = STATE_QUERY_ADDR;
            cu->cachable = SESS_UNCACHABLE;

            chk = push_sess(think_queue, cu);
            if(chk) {
Trace("timeout waited friend sess# %d from wait Q, fail to push think Q\n",
                    cu->sid);
                d++;

                DoSendMessage(cu);
                DoCloseSess(cu);
                DoReleaseSess(cu);
                cu = NULL;

            }
            else {
Trace("timeout waited friend sess# %d from wait Q, push think Q\n",
                    cu->sid);
                w++;
            }
        }
        else
        if(cu->cache_entry &&
            ((skcache_entry_t*)cu->cache_entry)->oid ==
                ((skcache_entry_t*)this->cache_entry)->oid) {

            /* re-push to think Q */
            chk = push_sess(think_queue, cu);
            if(chk) {
Trace("pick up waited friend sess# %d from wait Q, fail to push think Q\n",
                    cu->sid);
                d++;

                DoSendMessage(cu);
                DoCloseSess(cu);
                DoReleaseSess(cu);
                cu = NULL;

            }
            else {
Trace("pick up waited friend sess# %d from wait Q, push think Q bk# %d,\n",
                    cu->sid, cu->fhash);
                w++;
            }
            cu = NULL;
        }
        else {
#if 0
            chk = push_sess(wait_queue, cu);
#endif
            chk = sq_timedpush(wait_queue, cu, 5000);
            if(chk) {
                DoSendMessage(cu);
                DoCloseSess(cu);
                DoReleaseSess(cu);

            }
            cu = NULL;
        }
        i++;
    }

    MUTEX_UNLOCK(&wait_check_lock);

    Trace("check_friend_lock: sess# %d, oid# %d, n %d, w %d, d %d\n",
        this->sid, ((skcache_entry_t*)this->cache_entry)->oid,
        n, w, d);

    return w;
}

/*
 * release_buf - release allocated area
 *
 */
int
release_buf(sess_t *this)
{
    int ret;
    int chk;
    bucket_t *bk;
    skcache_entry_t *centry;


#ifdef TRACE_CACHE
    Trace("release_buf: sess# %d, '%s' [%d:%d] %s\n",
        this->sid, this->loc, this->fhash, this->shash,
        MSG_STATE_[this->state]);

#if 0
    cache_showblockaddr(this->fhash, 0);
#endif

#endif

    bk = &skcache_bucket_base[this->fhash];

    LOCK_BK(bk);


    centry = this->cache_entry;
    if(centry) {
        centry->ref--;

#if 0
        if(centry->owner==this->sid) {
            centry->owner = 0;
        }
#endif
        if(centry->owner==this) {
            centry->owner = NULL;
        }

        if(centry->ref>0) {
#ifdef USE_WQUEUE
            check_friend_lock(this);
#endif
        }

        /* forget entry */
        this->cache_entry = NULL;
    }

    ret = 0;

 assert(bk->cup.num==bk->entry_use);
    UNLOCK_BK(bk);
    bk = NULL;

out:

    return ret;
}



int
DoAccept(int listen_fd)
{
    struct sockaddr_in from;
    int len;
    int tmpfd;


    len = sizeof(from);

    MUTEX_LOCK(&accept_lock);
    tmpfd = accept(listen_fd, (struct sockaddr*)&from, &len);
    MUTEX_UNLOCK(&accept_lock);

    if(tmpfd<0) {
#if 1
        Error("accept error (%d)", errno);
#endif
        tmpfd = -1;
        goto done;
    }

    MUTEX_LOCK(&stat_lock);
    receive_request_count++;
    request_sirial_count++;
    MUTEX_UNLOCK(&stat_lock);

done:

    return tmpfd;
}


int
DoReadReq(sess_t *this)
{
    int nread;
    int c;
    int method;
    char url[L_URL];
    char version[L_URL];
    void *cpos;
    int chk;
    int err;
    int ims;

#if 0
    Trace("DoReadReq : sess# %d, cfd %d\n", this->sid, this->cfd);
#endif

#if 0
dump_stringsegment("before", this->reqbuf, this->reqlen);
#endif

try_read:
    nread = read(this->cfd,
                &this->reqbuf[this->reqlen], L_REQBUF-this->reqlen);
    err = errno;
    if(nread<0) {
        if(err==EINTR) {
            goto try_read;
        }
        Error("sess# %d, read() fail (%d)", this->sid, err);
        DoHalfClientCloseSess(this);

        this->reason = R_FAIL_CLIENT_READ;
        this->state = STATE_MESSAGE;

        MUTEX_LOCK(&stat_lock);
        ignore_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        return -1;
    }
    if(nread==0) {
        Error("sess# %d, read() empty, fd# %d", this->sid, this->cfd);
#if 0
        Error("sess# %d, th %x, fd# %d, empty client request (%d)",
            this->sid, pthread_self(), this->cfd, err);
#endif
        DoHalfClientCloseSess(this);

        this->reason = R_FAIL_CLIENT_READ_EMPTY;
        this->state = STATE_MESSAGE;

        MUTEX_LOCK(&stat_lock);
        ignore_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        return -1;
    }

    this->reqlen += nread;
    this->cread += nread;
    this->loc[0] = '\0';
#if 0
    this->last_action = time(NULL);
#endif

#if 0
dump_stringsegment("o.req.", this->reqbuf, this->reqlen);
#endif

    c = HTTP_find_header_term(this->reqbuf, this->reqlen);

#if 0
Trace("DoReadReq: sess %p, read %d, header term %d\n", this->sid, nread, c);
#endif

    if(c>0) {
#if 0
        dump_stringsegment("request", this->reqbuf, this->reqlen);
#endif

#if 0
 Trace(" 1 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif

        version[0] = '\0';
        ims = -1;

        c = HTTP_parse_request_header(this->reqbuf,
                &method, url, version,
                &ims, &this->cachable, &this->reload, &this->dhr_state);

        if(c) {
            Error("ignore request from fd #%d", this->cfd);
#if 1
            dump_stringsegment("request", this->reqbuf, this->reqlen);
#endif

            if(c==-2) {
                this->reason = R_TOOLONG_URL;
            }
            else {
                this->reason = R_IGNORE_URL;
            }

            this->state = STATE_MESSAGE;
            return -1;
        }

        chk = -1;
        if(version[0]) {
            chk =
             HTTP_parse_version(version, &this->cver_major, &this->cver_minor);
        }
        if(chk<0) {
            this->cver_major = 0;
            this->cver_minor = 9;
        }
#if 0
        Trace("sess# %d, chk %d, client HTTP ver %d.%d\n",
            this->sid, chk, this->cver_major, this->cver_minor);
#endif

#ifdef TRACE_SESS
        Trace("sess# %d, <client> method <%s>,URL '%s' version <%s>\n",
            this->sid, MSG_M_[method], url, version);
#endif

        this->method = method;

        strcpy(this->loc, url);
        calc_hash(this->loc,
            &this->fhash, &this->shash, &this->thash, &this->ffhash);

#if 0
Trace("rreq: %03d:%03d:%03d:%03d %s\n",
    this->fhash, this->shash, this->thash, this->ffhash, this->loc);
#endif


/*****
 ***** check cachable or not
 *****/

        if(istempURL(url)) {
            this->cachable  = SESS_UNCACHABLE;
            this->reload    = SESS_RELOAD;
        }

        if(ims>0) {
            this->ims = ims;
        }

#if 0
 Trace(" 3 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif

done_cachable_check:
#if 0
 Trace(" 4 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif

        this->state = STATE_THINK;
    }
    else {
        Error("DoReadReq: HTTP_find_header_term() returns %d", c);
#if 1
        dump_stringsegment("e.req", this->reqbuf, this->reqlen);
#endif
    }
#if 0
 Trace(" 5 sfd# %d, cfd# %d\n", this->sfd, this->cfd);
#endif

    return 0;
}


int
DoSetReq(sess_t *this, char *aurl)
{
    int nread;
    int c;
    int method;
    char url[L_URL], version[L_URL];
    char protocol[L_URL], portstr[L_URL];
    void *cpos;
    int istmp;


#if 0
    Trace("DoSetReq : cfd %d\n", this->cfd);
#endif

    this->reqlen = sprintf(this->reqbuf,
            "GET %s HTTP/1.0\r\nAccept: */*\r\n\r\n", aurl);

    this->loc[0] = '\0';

#if 0
    this->last_action = time(NULL);
#endif

    c = HTTP_parse_simple_request(this->reqbuf, &method, url, version);
    if(c) {
        Error("ignore request from fd #%d", this->cfd);
#if 1
        dump_stringsegment("request", this->reqbuf, this->reqlen);
#endif

        this->reason = R_IGNORE_URL;
        this->state = STATE_MESSAGE;
        return -1;
    }

#if 0
    Trace("receive request\n");
    Trace("  method  <%s>\n  URL     <%s>\n  version <%s>\n",
        method, url, version);
#endif

    this->method = method;

    strcpy(this->loc, url);
    calc_hash(this->loc,
        &this->fhash, &this->shash, &this->thash, &this->ffhash);

#if 0
Trace("sreq: %03d:%03d:%03d:%03d %s\n",
    this->fhash, this->shash, this->thash, this->ffhash, this->loc);
#endif


    this->state = STATE_THINK;

    return 0;
}



int
DoReadReqFirst(sess_t *this)
{
    int ret;
    int b_state;


    b_state = this->state;
    this->state = STATE_RECV_REQ;
    ret = DoReadReq(this);
#if 0
    Trace("DoReadReqFirst: sess# %d, %s -> %s , ret %d\n",
        this->sid, MSG_STATE_[b_state], MSG_STATE_[this->state], ret);
#endif

    return ret;
}


int
DoReadReqRest(sess_t *this)
{
    int ret;
    int b_state;


    b_state = this->state;
    ret = DoReadReq(this);
#if 0
    Trace("DoReadReqRest: sess# %d, %s -> %s , ret %d\n",
        this->sid, MSG_STATE_[b_state], MSG_STATE_[this->state], ret);
#endif

    return ret;
}

int
HostnameNormalize(char *src)
{
    register char *p;
    register int c;
    int d;

    p = src;
    c = 0;
    while(*p) {
        c++;
        if(*p=='.') {
            goto have_dot;
        }
        p++;
    }

not_have_dot:

    MUTEX_LOCK(&config_lock);
    if(c+strlen(localdomain_padding)+1<L_HOST) {
        strcat(p, localdomain_padding);
    }
    MUTEX_UNLOCK(&config_lock);

    return 0;

have_dot:
    return 0;
}


#if 0
int check_bkentry(bucket_t*,char*);
#endif


extern int skcache_mem_keep_min;
extern int skcache_mem_keep_cur;
extern int skcache_mem_keep_max;
extern int skcache_entry_keep_min;
extern int skcache_entry_keep_cur;
extern int skcache_entry_keep_max;


int
scan_frags_in_bucket(bucket_t *bk)
{
    skiplist_t *list;
    skiplist_cell_t *x;
    skcache_entry_t *ent;
    skcache_frag_t *cur;
    int in_entry;
    int in_bucket;

    list = &bk->cup;
    x = list->header->skip[0];

    in_bucket = 0;

    while(x!=list->term) {
        ent = (skcache_entry_t*) x->entity;
        cur = ent->m_recorded;

        in_entry = 0;

        while(cur) {
            if(cur->next==cur) {
                Error("bk#%d, entity %p, %d th frag %p is looped itself",
                    bk->bid, ent, in_entry, cur);
                ent->cracked = 1;
                break;
            }
            in_entry++;

            cur = cur->next;
        }
        in_bucket += in_entry;
        
        x = x->skip[0];
        if(x==NULL)
            break;
    }

#if 0
    Trace("bk# %d, has %d frag\n", bk->bid, in_bucket);
#endif

    return in_bucket;
}



int
comp_targets(const void *xa, const void *xb)
{
    skcache_gc_target_t *a = (skcache_gc_target_t*) xa;
    skcache_gc_target_t *b = (skcache_gc_target_t*) xb;

    return a->hot - b->hot;
}

#ifndef TMP_NTARGET
#define TMP_NTARGET (2048)
#endif

int
reaper(bucket_t *bk, int nq)
{
    int ret;
    int chk;
    int csum;
    int action;
    int ntry;
    int new_keep, last_keep;
    int i;
    int cp;
    long long bsum;
    int act_sum[5];

    skiplist_t *list;
    skiplist_cell_t *x, *n;
    skiplist_cell_t *last;
    skiplist_cell_t *update[SKL_MAXLEVEL];

    skcache_entry_t *ent;
    struct timeval xbefore, xafter;
    time_t now;

    int checkpoint;

    int target_hot;
    skcache_entry_t *target_pos;

    int tmp_ntarget;
    skcache_gc_target_t tmp_targets[TMP_NTARGET];

    int c;

    int max_cp;
    int wait_flag;

    char* skcache_label4_ref(void *pp);



    /* NOTE:
     *  bucket is already locked. don't re-lock.
     */

    gettimeofday(&xbefore, NULL);
    now = xbefore.tv_sec;

    if(nq<0) {
        nq = MAX_GC_NTARGET;
    }


    MUTEX_LOCK(&stat_lock);
    new_keep = skcache_mem_keep_cur;
    MUTEX_UNLOCK(&stat_lock);

#ifdef TRACE_REAPER
#ifndef NDEBUG
    {
    c = scan_frags_in_bucket(bk);
    Trace("reaper: before bk# %d, frag bk->frag_use %d, actual %d\n",
        bk->bid, bk->frag_use, c);
    }
#endif
#endif


#ifdef TRACE_REAPER
    Trace(
"reaper: bk# %d, nq %d, time %d %d, target %d, elem %4d/%d, entry %d+%d %d, frag %d\n",
        bk->bid, nq,
        xbefore.tv_sec, bk->gc_last,
        bk->gc_ntarget,
        bk->cup.num, bk->elemnum,
        bk->entry_use, bk->entry_free, bk->entry_all,
        bk->frag_use);

#endif

assert(bk->cup.num==bk->entry_use);

    ret = 0;

    target_hot = now;
    target_pos = NULL;

    if(bk->gc_ntarget>0 && xbefore.tv_sec < bk->gc_last) {
#ifdef TRACE_REAPER
        Trace("  skip scanning bk# %d, ntarget %d (hot %d sec)\n",
            bk->bid, bk->gc_ntarget, bk->gc_last - xbefore.tv_sec);
#endif
        goto delete_targets;
    }


    /***
     ***
     ***/

    MUTEX_LOCK(&stat_lock);
    msweeper_count++;
    MUTEX_UNLOCK(&stat_lock);


    list = &bk->cup;

#if 1
    /* clear headers */
    list->level = 0;
    x = list->header;
    x->entity = list->term;
    for(i=1;i<=list->maxlevel;i++) {
        x->skip[i] = list->term;
    }
#endif

    ntry = 0;

try_sweep:
    ntry++;

assert(bk->cup.num==bk->entry_use);
#ifdef TRACE_REAPER
    Trace("  try %d, entries %d (%d,%d), keep %d sec, global %d sec \n",
        ntry, bk->entry_use, bk->entry_hwater, bk->entry_lwater,
        new_keep, skcache_mem_keep_cur);
#endif

    csum = 0;
    bsum = 0LL;
    cp = 0;

    max_cp = bk->entry_use;

    x = list->header;
    x = x->skip[0];
    last = NULL;

    for(i=0;i<5;i++) {
        act_sum[i] = 0;
    }

    tmp_ntarget = 0;

    while(x!=list->term) {
assert(bk->cup.num==bk->entry_use);

        if(cp>=max_cp) {
            Error("cp %d is too big than max_cp %d", cp, max_cp);
        }

        cp++;
        n = x->skip[0];
        ent = x->entity;
#ifndef NDEBUG
        if(ent==NULL) {
            Error("bk# %d, %p/ %d th %x has empty elem.",
                bk->bid, bk, cp);
            goto next_elem;
        }
#endif
#ifdef CHECK_MAGIC
        if(ent->magic[0]!=MAGIC_0|| ent->magic[1]!=MAGIC_1) {
            Error("reaper: ignore magic bk# %d, ent %p", bk->bid, ent);
            goto next_elem;
        }
#endif

#if 1
        for(i=1;i<=list->maxlevel;i++) {
            x->skip[i] = list->term;
        }
#endif

        action = skcache_dicision_w_th(bk, ent, 0,
            xbefore.tv_sec,
            skcache_mem_keep_min, new_keep,
            skcache_entry_keep_cur);

        act_sum[action]++;

#if 0
 Trace("  %3d: %p: entity %#x action %d\n", cp, x, ent, action);
#endif
        switch(action) {
        case 0: /* KEEP */
            csum++;
            bsum += ent->totallength;

            if(ent->hot>checkpoint) {
                checkpoint = ent->hot;
            }

            if(ent->hot<target_hot) {
                target_hot = ent->hot;
                target_pos = ent;
            }

            break;

        case 1: /* SWAPOUT */
        case 3: /* SWAPOUT + PURGE */
            /* mark */

            if(tmp_ntarget<TMP_NTARGET) {
#if 0
Trace("store  bk# %d, %d/%-2d : pos %p oid# %d\n",
    bk->bid, tmp_ntarget, TMP_NTARGET,  ent, ent->oid);
#endif
                tmp_targets[tmp_ntarget].pos = ent;
                tmp_targets[tmp_ntarget].hot = ent->hot;
                tmp_targets[tmp_ntarget].oid = ent->oid;
                tmp_ntarget++;
            }


            break;

        case 2: /* PURGE */
purge_elem:
            /* frag level */
            skcache_entry_purge(bk, ent);

            /* entry level */
            skcache_push_idle_entry(bk, ent);
            ent = NULL;
            x->entity = NULL;

#if 0
            /* elem level */
            for(i=1;i<=list->maxlevel;i++) {
                x->skip[i] = list->term;
            }
#endif

            /* release from skiplist */
            if(last==NULL) {
                list->header->skip[0] = n;
            }
            else{
                last->skip[0] = n;
            }

            x->skip[0] = bk->elem_freelist;
            bk->elem_freelist = x;
            list->num--;

#ifdef TRACE_REAPER
#ifndef NDEBUG
 {
    skiplist_cell_t *y;
    int m, k;

    y = list->header;
    y = y->skip[0];
    m = 0;
    while(y!=list->term) {
        if(y->entity==NULL) {
            Trace("re:ch: bk# %d, %3d th entity is empty\n",
				bk->bid, m);
        }
        for(k=0;k<=list->maxlevel;k++) {
            if(y->skip[k]==NULL) {
                Trace("re:ch: bk# %d, %3d th entry, %d th skip is NULL\n",
					bk->bid, m, k);
            }
        }

        y = y->skip[0];
        m++;
    }
 }
#endif
#endif

            break;
        }

        if(action!=2) {
            last = x;
        }

next_elem:
        x = n;
    }

#if 0
  Trace("  use %d, count %d, keeptime %d, hold %d entries %lld bytes (%.2fMB)\n\
  action bk# %d, keep %d, swap %d, purge %d, swap+purge %d, markd %d\n",
        bk->entry_use, cp, new_keep,
        csum, bsum, fdM(bsum),
        bk->bid,
        act_sum[0], act_sum[1], act_sum[2], act_sum[3], act_sum[4]);
#endif

#if 1
    if(cp!=(act_sum[0]+act_sum[1]+act_sum[2]+act_sum[3]+act_sum[4])) {
        Error("strange A bk# %d, (sum of action %d, cp %d)",
            bk->bid,
            act_sum[0]+act_sum[1]+act_sum[2]+act_sum[3]+act_sum[4], cp);
    }
#endif


    if(tmp_ntarget>0) {
#if 0
        Trace("  bk# %d, tmp_ntarget %d\n", bk->bid, tmp_ntarget);
#endif
        qsort(tmp_targets, tmp_ntarget,
            sizeof(skcache_gc_target_t), comp_targets);

        bk->gc_rest = tmp_ntarget;
        
        for(i=0;i<MIN(tmp_ntarget, MAX_GC_NTARGET);i++) {
            bk->gc_targets[i] = tmp_targets[i];
        }
        bk->gc_ntarget = i;

#if 0
        for(i=0;i<bk->gc_ntarget;i++) {
            Trace("store  bk# %d, %d/%-2d : pos %p oid# %d\n",
                bk->bid, i, MAX_GC_NTARGET, 
                bk->gc_targets[i].pos, bk->gc_targets[i].oid);
        }
#endif

    }
#if 1   /* cleaning still low water */
    else
    if(ntry<5 && bk->entry_use > bk->entry_lwater) {
        last_keep = new_keep;
        new_keep = 9*new_keep/10;

#ifdef TRACE_REAPER
        Trace("  retry<%d>, bk# %d, %d/%d new keep %d\n",
            ntry, bk->bid, bk->entry_use, bk->entry_lwater, new_keep);
#endif
        goto try_sweep;
    }
    else{
#ifdef TRACE_REAPER
        Trace("  giveup<%d>, bk# %d, %d/%d\n",
            ntry, bk->bid, bk->entry_use, bk->entry_lwater);
#endif
        bk->gc_rest = 0;
        bk->gc_ntarget = 0;
    }
#else
    else{
        bk->gc_rest = 0;
        bk->gc_ntarget = 0;
    }
#endif

    bk->gc_last = xbefore.tv_sec+5;

delete_targets:
 {
 
    c = 0;

    if(bk->gc_ntarget>0) {
        do {
#if 0
            Trace("remove bk# %d, %d/%-2d : pos %p, oid# %d // actual oid# %d\n",
                bk->bid, c, nq, 
                bk->gc_targets[0].pos,
                bk->gc_targets[0].oid, 
                bk->gc_targets[0].pos->oid);
#endif

            if(bk->gc_targets[0].pos
              && bk->gc_targets[0].pos->hot == bk->gc_targets[0].hot
              && bk->gc_targets[0].pos->oid == bk->gc_targets[0].oid) {

                chk = skcache_entry_swapout(bk, bk->gc_targets[0].pos);
                if(chk<0) {
#if 0
                    Trace("    skip reason %d\n", chk);
#endif
                }
                else {
 if(chk) {
    Error(" 1 fail swapout ent %p, ret %d", bk->gc_targets[0].pos, chk);
 }
                    (bk->gc_targets[0].pos)->ref = 0;   /* force */
                    chk = skcache_entry_delete(bk, bk->gc_targets[0].pos);
 if(chk) {
    Error(" 1 fail delete ent %p, ret %d", bk->gc_targets[0].pos, chk);
 }
                    bk->gc_targets[0].pos = NULL;
                }
                c++;
            }

            memmove(&bk->gc_targets[0], &bk->gc_targets[1],
                (bk->gc_ntarget-1)*sizeof(skcache_gc_target_t));
            bk->gc_ntarget--;

        } while(bk->gc_ntarget>0 && c<nq);

        MUTEX_LOCK(&stat_lock);
        compaction_count += c;
        MUTEX_UNLOCK(&stat_lock);

        bk->gc_checkpoint = bk->gc_targets[0].hot;

#ifdef TRACE_REAPER
        Trace("  bk# %d, LRU remove %d/%d, rest %d\n",
            bk->bid, c, nq, bk->gc_ntarget);
#endif
    }
    else
    if(target_pos) {
        chk = skcache_entry_swapout(bk, target_pos);
        if(chk<0) {
#if 0
            Trace("    skip reason %d\n", chk);
#endif
        }
        else {
 if(chk) {
    Error(" 2 fail swapout ent %p, ret %d", target_pos, chk);
 }
            target_pos->ref = 0; /* force */
            chk = skcache_entry_delete(bk, target_pos);
 if(chk) {
    Error(" 2 fail delete ent %p, ret %d", target_pos, chk);
 }
            target_pos = NULL;
            c++;
        }

        MUTEX_LOCK(&stat_lock);
        compaction_count++;
        MUTEX_UNLOCK(&stat_lock);

        bk->gc_checkpoint = target_hot;
        target_hot = -1;

#ifdef TRACE_REAPER
        Trace("  bk# %d, NRU remove %d/%d, rest %d\n",
            bk->bid, c, nq, bk->gc_ntarget);
#endif
    }
    else {
        bk->gc_checkpoint = -1;

#if 0
        Trace("  bk# %d, none\n",
            bk->bid);
#endif
    }

#if 0
    Trace("  bk# %d, remove %d/%d, rest %d\n",
        bk->bid, c, nq, bk->gc_ntarget);
#endif
 }

send_mark:

out:
    gettimeofday(&xafter, NULL);
    sim_stat_add_value(&msweep_time, KOTETU_SUB_TIME(xafter, xbefore));
    sim_stat_add_value(&msweep_time_short, KOTETU_SUB_TIME(xafter, xbefore));

#ifdef TRACE_REAPER
    Trace("  bk# %d, checkpoint %d <%d sec ago>, takes %d msec\n",
        bk->bid, bk->gc_checkpoint, now - bk->gc_checkpoint,
        KOTETU_SUB_TIME(xafter, xbefore));

#ifdef NDEBUG
    {
    c = scan_frags_in_bucket(bk);
    Trace("reaper: after  bk# %d, frag bk->frag_use %d, actual %d\n",
        bk->bid, bk->frag_use, c);
    }
#endif
#endif

    return ret;
}
/* end of reaper */




int
DoThink(sess_t *this)
{
    int ihave;
    int chk;
    char protocol[L_URL], portstr[L_URL];

    bucket_t *bk;
    skcache_entry_t *stored_pos;
    int check_already_have;
    skcache_entry_t *newone;
    char msg[L_PATH];

    int sig_pos;
    int dcount; 
    int dtime;

    struct timeval xbefore, xafter;
    extern int sq_check_count();


    chk = ParseURL(this->loc, protocol,
            this->servername, portstr, this->serverpath);
    if(portstr[0]=='\0') {
        this->serverport = 80;
    }
    else {
        this->serverport = atoi(portstr);
    }
    chk = HostnameNormalize(this->servername);

    switch(this->dhr_state){
      int d1, d2;
    case 'L':
      d1 = dhr_hash1(this->servername, this->serverpath, n_proxy);
      d2 = dhr_hash2(this->servername, this->serverpath, n_proxy);
      if((d1 == proxy_id) || (d2 == proxy_id)){
	if(d1 == proxy_id){
	  this->dhr_hash = d1;
	  this->dhr_hash_tmp = d2;
	  this->dhr_state = 'A';
	}
	else{
	  this->dhr_hash = d2;
	  this->dhr_hash_tmp = d1;
	  this->dhr_state = 'B';
	}
      }
      else{
	this->cachable = SESS_UNCACHABLE;
	this->state = STATE_QUERY_ADDR;
	if(random()%2){
	  this->dhr_hash = d1;
	  this->dhr_state = 'A';
	}
	else{
	  this->dhr_hash = d2;
	  this->dhr_state = 'B';
	}
	return 0;
      }
      break;
    case 'A':
      this->dhr_hash = dhr_hash2(this->servername, this->serverpath, n_proxy);
      this->dhr_state = 'M';
      break;
    case 'B':
      this->dhr_hash = dhr_hash1(this->servername, this->serverpath, n_proxy);
      this->dhr_state = 'M';
      break;
    case 'S':
      this->dhr_state = 'E';
      this->dhr_hash = -1;
      break;
    case 'M':
      this->rcode = 306;
      this->dhr_hash = -1;
      this->dhr_state = 'R';
      break;
    case 'E':
      this->dhr_hash = -1;
      break;
    }

#if 0 /*for HR */
	if((this->dhr_hash = 
            dhr_hash1(this->servername, this->serverpath, n_proxy))
	   != proxy_id){
	  this->cachable = SESS_UNCACHABLE;
	}
#endif


#if 0
    Trace("DoThink: '%s' [%04d:%04d:%04d] '%s' %s %s\n",
        this->loc, this->fhash, this->shash, this->thash, this->loc,
        this->cachable==SESS_CACHABLE ? "cachable" : "no-cachable",
        this->reload==SESS_RELOAD ? "reload" : "not-reload");
#endif

    /* this system permit only GET and POST method */
    if(! (this->method == M_GET || this->method == M_POST) ) {
        Trace("ignore method, sess %p (#%d)\n", this, this->sid);

        MUTEX_LOCK(&stat_lock);
        ignore_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->reason = R_UNSUPPORT_METHOD;
        this->state = STATE_MESSAGE;
        return 0;
    }

#if 1
    if(strncasecmp(this->loc, "http:", 5)!=0) {
        this->reason = R_UNSUPPORT_PROTOCOL;
        this->state = STATE_MESSAGE;
        return 0;
    }
#endif

    if(strcmp(this->servername, report_hook_hostname)==0)
    {
#if 0
        Trace("report request. sess# %d %p\n", this->sid, this);
#endif

        MUTEX_LOCK(&stat_lock);
        report_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->state = STATE_REPORT;
        this->cachable = SESS_UNCACHABLE;
        this->reload   = SESS_RELOAD;
        return 0;
    }


    sim_stat_add_value(&url_first_hash_stat,  this->fhash);
    sim_stat_add_value(&url_second_hash_stat, this->shash);
    sim_stat_add_value(&url_third_hash_stat,  this->thash);
    sim_stat_add_value(&url_fourth_hash_stat,  this->ffhash);

    sim_stat_add_value(&req_field_type_stat, 
        this->reload + this->cachable + this->publish);

    switch(this->method) {
    /* GET method */
    case M_GET:

get_first:

        bk = &skcache_bucket_base[this->fhash];
        LOCK_BK(bk);

 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());

#if 0
    Trace("think: * sess# %d, bk# %d, %p\n",
        this->sid, bk->bid, bk);
#endif


 assert(check_emptyentity(bk));

try_get_newone:
 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
    newone = skcache_pull_new_entry(bk);
 assert(check_bkinfo());

    if(newone==NULL) {
        Error("think: sess# %d, no free entry element in bk# %d, %p <2b>",
            this->sid, this->fhash, bk);

        this->state  = STATE_MESSAGE;
        this->reason = R_MEMORY_OVERFLOW;

        goto get_last;
    }

 assert(bk->cup.num==bk->entry_use-1);
 assert(check_bkinfo());

#if 0
    Trace(
"request %p sess# %-7d,      %03d:%03d:%03d:%03d %s\n",
    this, this->sid,
    this->fhash, this->shash, this->thash, this->ffhash, this->loc);
#endif
#if 0
    Trace("sess# %d, TH magic %02x %02x\n",
            this->sid, newone->magic[0], newone->magic[1]);
#endif

    newone->oid = -1;
    newone->h1 = this->fhash;
    newone->h2 = this->shash;
    newone->h3 = this->thash;
    newone->h4 = this->ffhash;
    strcpy(newone->name, this->loc);

#if 0
    Trace(
"newone %p oid# %-7d, ref %d, %03d:%03d:%03d:%03d %s\n",
        newone, newone->oid, newone->ref,
        newone->h1, newone->h2, newone->h3, newone->h4,
        newone->name);
#endif

    gettimeofday(&xbefore, NULL);
    stored_pos =
        (skcache_entry_t*) sk_assign(&bk->cup, newone,
            &bk->elem_freelist, skcache_entry_cmp,
            &check_already_have, 0);
    gettimeofday(&xafter, NULL);
#if 0
    Trace("th: sess# %d, stored_pos %p, check_already_have %d\n",
            this->sid, stored_pos, check_already_have);
#endif
 assert(check_bkinfo());
 assert(check_emptyentity(bk));

    if(stored_pos==NULL || check_already_have<0) {
        Error("think: cannot assign to entry free list of bk# %d, <1b>",
                this->fhash);
        Trace("think: stored_pos %p, check_already_have %d\n",
            stored_pos, check_already_have);

        this->state  = STATE_MESSAGE;
        this->reason = R_UNKNOW_ERROR;

        skcache_push_idle_entry(bk, newone);
#if 0
/* XXX */
        break;
#endif

        goto get_last;
    }

    /* NOTE:
     * stored_pos != NULL
     */
    if(stored_pos->ref<0) {
        Error("th: dead object **STRANGE** entry %p", stored_pos);
    }

    if(stored_pos!=newone) {
        /* HIT */
        /* 'newone' is not stored in bucket, becasue same old entry
         * is stored alread.  use old one */
        skcache_push_idle_entry(bk, newone);
        newone = NULL;
        strcpy(msg, "HIT ");
    }
    else {
        /* MISS */
        /* 'newone' is stored in bucket */
        newone = NULL;
        strcpy(msg, "MISS");
    }
#if 0
    Trace(
"%s   %p oid# %d, ref %d, m %p, %03d:%03d:%03d:%03d %s\n",
        msg,
        stored_pos, stored_pos->oid, stored_pos->ref,
        stored_pos->m_recorded,
        stored_pos->h1, stored_pos->h2, stored_pos->h3, stored_pos->h4,
        stored_pos->name);
#endif

 assert(bk->cup.num==bk->entry_use);

    /* not found in memory
     * the object is allocated new area
     */
    if(check_already_have==0 && stored_pos->m_recorded==NULL) {
        int xb, xa;
        int sig;

        sig_pos = DSIG_POS(stored_pos->h1, stored_pos->h2, stored_pos->h3);

        MUTEX_LOCK(&dir_mutex);
        sig     = DSIG_ISSET(sig_pos, stored_pos->h4/DSIG_VALUE_RES);
        dcount  = dir_count[sig_pos];
        dtime   = dir_time[sig_pos];
        MUTEX_UNLOCK(&dir_mutex);

#if 0
        Trace(
        "dir# %d [ %3d %3d %3d %4d ] time %4d, count %4d, sig %8x => %s\n",
            sig_pos,
            stored_pos->h1, stored_pos->h2, stored_pos->h3, stored_pos->h4,
            time(NULL) - dtime, dcount, sig,
            sig ? "MATCH" : "nomatch"
        );
#endif
#if 0
        Trace("sess# %d, dir# %d, count %d, time %d, h4 %d => %d -> %d\n",
            this->sid, sig_pos, dcount, dtime,
            stored_pos->h4, stored_pos->h4/DSIG_VALUE_RES, sig);
        dsig_show(&dir_sig[sig_pos*DSIG_ELEM_WIDTH]);
#endif

        if(sig)
        {

            xb = bk->entry_use;
 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());

            chk = skcache_entry_swapload(bk, stored_pos);
            if(chk==0) {
                check_already_have=1;
#if 0
                Trace("  got entry oid# %d via swapload\n",
                    stored_pos->oid);
#endif
            }

 assert(bk->cup.num==bk->entry_use);
 assert(check_bkinfo());
            xa = bk->entry_use;
#if 0
Trace("= %d %d : %d -> %d /%d\n", chk, xa-xb, xb, xa, bk->entry_all);
#endif
        }
    }

#ifdef CHECK_MAGIC
    if(stored_pos->magic[0]!=MAGIC_0|| stored_pos->magic[1]!=MAGIC_1) {
        Error("think: ignore magic bk# %d, stored_pos %p, sess# %d",
            bk->bid, stored_pos, this->sid);
    }
#endif


    this->cache_entry = stored_pos;
    stored_pos->ref++;
    stored_pos->hot = xafter.tv_sec;
 assert(CHKTIME(stored_pos->hot));

#if 0
Trace("DoThink: sess# %d, cache_entry %p, owner %p, ref %d\n",
        this->sid, stored_pos, stored_pos->owner, stored_pos->ref);
#endif
#if 0
Trace("DoThink: sess# %d, entry %p, oid# %d, %s(%d), owner %p, ref %d\n",
        this->sid, stored_pos, stored_pos->oid,
        (check_already_have==1) ? "USED" :
            ((check_already_have== -1) ? "ERROR" : "NEW"),
        check_already_have,
        stored_pos->owner, stored_pos->ref);
#endif

 assert(check_bkinfo());

#if 1
    if(check_already_have==0) {
    }
    else {
        /* error check */
        if(stored_pos->oid<0) {
            Trace("negative memory entry %p oid# %d, ref %d <%d>\n",
                stored_pos, stored_pos->oid, stored_pos->ref,
                check_already_have);
        }
    }
#endif

    if(this->reload!=SESS_RELOAD && this->cachable==SESS_CACHABLE) {

        if(check_already_have==0) {
#if 0
Trace("  *UNKNOWN*\n");
#endif

        if(this->dhr_state == 'R'){
	  skcache_entry_delete(bk, stored_pos);
	  UNLOCK_BK(bk);
	  this->reason = R_DHR_SECONDARY_MISS;
	  this->state = STATE_DHR_SECONDARY_MISS;
	  return 0;
	}


            /* it is new object, initialize attributes */
#if 0
            stored_pos->owner       = this->sid;
#endif
            stored_pos->owner       = this;
            stored_pos->oid         = this->sid;
            stored_pos->issue_time  = this->last_action;

#if 0
            strcpy(stored_pos->name, this->loc);
#endif

            stored_pos->totallength = (off_t) 0L;
            stored_pos->m_recorded = NULL;
            stored_pos->d_recorded = -1;


#if 0
Trace("  stored_pos        %p m_recorded %p\n",
        stored_pos, stored_pos->m_recorded);
Trace("  this->cache_entry %p m_recorded %p\n",
        this-> cache_entry, ((skcache_entry_t*)this->cache_entry)->m_recorded);
#endif

            MUTEX_LOCK(&stat_lock);
            miss_reply_count++;
            MUTEX_UNLOCK(&stat_lock);

            this->state = STATE_QUERY_ADDR;

        }
        else {
#if 0
Trace("  *KNOWN*\n");
#endif
#if 0
            Trace("known object: request sess# %d, %d:%d:%d:%d %s\n\
found %p oid# %d, ref %d, %d:%d:%d:%d %s\n",
    this->sid,
    this->fhash, this->shash, this->thash, this->ffhash, this->loc,
    stored_pos, stored_pos->oid, stored_pos->ref,
    stored_pos->h1, stored_pos->h2, stored_pos->h3, stored_pos->h4,
    stored_pos->name);

#endif

            if(stored_pos->ref>1) {
                this->state = STATE_WAIT_FRIEND;

                MUTEX_LOCK(&stat_lock);
                nfriend++;
                friend_count++;
                MUTEX_UNLOCK(&stat_lock);

#if 0
            Trace("sess# %d, wait friend session with %p %p\n",
                this->sid, this->cache_entry, stored_pos);

            Trace("stored %p oid# %4d ref %3d, %4d %4d %4d %4d %s\n",
                stored_pos, stored_pos->oid, stored_pos->ref,
                stored_pos->h1, stored_pos->h2,
                stored_pos->h3, stored_pos->h4, stored_pos->name);
#endif

            }
            else
            if(stored_pos->owner==this) {
#ifdef TRACE_SESS
                Trace("sess# %d, cachable, not have [%d:%d]\n",
                    this->sid, this->fhash, this->shash);
#endif

                MUTEX_LOCK(&stat_lock);
                miss_reply_count++;
                MUTEX_UNLOCK(&stat_lock);

                this->state = STATE_QUERY_ADDR;


            }
            else {

                if(stored_pos->cracked) {
                /* cracked object
                 * don't warry other session
                 */
#ifdef TRACE_SESS
                    Trace("sess# %d, cachable, have and cracked [%d:%d]\n",
                        this->sid, this->fhash, this->shash);
#endif
                    MUTEX_LOCK(&stat_lock);
                    miss_reply_count++;
                    MUTEX_UNLOCK(&stat_lock);

                    this->state = STATE_QUERY_ADDR;

                }
                else
                if(stored_pos->owner==NULL) {
                    /* the object is not included other sessions
                     * this session can use the object
                     */
#ifdef TRACE_SESS
                    Trace("sess# %d, cachable, have [%d:%d], cold\n",
                        this->sid, this->fhash, this->shash);
#endif
#if 0
    cache_showblockaddr(this->fhash, 1);
#endif
#if 0
                    if(stored_pos->expire>0) {
                        Trace("sess# %d, expire %d, now %d\n",
                            this->sid, stored_pos->expire, xafter.tv_sec);
                    }
#endif
                    if(stored_pos->expire>0 &&
                        xafter.tv_sec > stored_pos->expire) {

                        MUTEX_LOCK(&stat_lock);
                        miss_reply_count++;
                        MUTEX_UNLOCK(&stat_lock);

                        this->state = STATE_QUERY_ADDR;

                    }
                    else {
                        MUTEX_LOCK(&stat_lock);
                        hit_reply_count++;
                        MUTEX_UNLOCK(&stat_lock);

                        this->state = STATE_SEND_CACHE;

                        this->ishit = SESS_HIT;

                        if(stored_pos) {
                            stored_pos->count++;
                            stored_pos->hot = this->last_action;
  assert(CHKTIME(stored_pos->hot));
                            this->count = stored_pos->count;
                            this->oid = stored_pos->oid;
#if 0
    if(this->count>1) {
        Trace("&# oid# %8d, %3d, %d, name %s\n",
        stored_pos->oid, stored_pos->count, stored_pos->hot, stored_pos->name);
    }
#endif
                        }

                    }
                }
                else {
                    /* the object is included another session
                     * this session have to wait complete of that
                     */
#ifdef TRACE_SESS
                    Trace("sess# %d, cachable, have [%d:%d], hot\n",
                        this->sid, this->fhash, this->shash);
#endif
#if 1
                    Trace("sess# %d, cachable, have [%d:%d], hot\n",
                        this->sid, this->fhash, this->shash);
#endif

                    /* not hit and not miss */

                    this->state = STATE_WAIT_FRIEND;
					this->hf++;

                    MUTEX_LOCK(&stat_lock);
                    nfriend++;
                    MUTEX_UNLOCK(&stat_lock);
#if 0
                    this->reason2 = stored_pos->owner;
#endif
#if 0
                    if(this->mode==MODE_CLIENT) {
                    Trace("sess# %d, to waiting owner %p, client\n",
                        this->sid, stored_pos->owner);
                    }
                    else {
                    Trace("sess# %d, to waiting owner %d, self\n",
                        this->sid, stored_pos->owner);
                    }
#endif
                }
            }
        }
    }
    else {
#if 0
        Trace("sess# %d, uncachable, make new entry\n", this->sid);
#endif

        if(check_already_have==1) {
#if 0
            Trace("th: uncahable, have entry %p, oid# %d, ref %d\n",
                stored_pos, stored_pos->oid, stored_pos->ref);
#endif

            if(stored_pos->owner==NULL) {

                /* the program have previous object but it has not owner.
                 * delete the object and assign new entry to session. */

                stored_pos->cracked = 1;
                MUTEX_LOCK(&stat_lock);
                miss_reply_count++;
                MUTEX_UNLOCK(&stat_lock);

                this->state = STATE_QUERY_ADDR;
            }
            else {
                this->state = STATE_WAIT_FRIEND;

                MUTEX_LOCK(&stat_lock);
                nfriend++;
                MUTEX_UNLOCK(&stat_lock);
#if 0
                this->reason2 = stored_pos->owner;
#endif
            }
        }
        else {
#if 0
            Trace("th: uncahable, not have entry %p, oid# %d, ref %d\n",
                stored_pos, stored_pos->oid, stored_pos->ref);
#endif

            this->oid = this->sid;

#if 0
            stored_pos->owner       = this->sid;
#endif
            stored_pos->owner       = this;
            stored_pos->oid         = this->sid;
            stored_pos->issue_time  = this->last_action;

            stored_pos->cracked = 1;
            MUTEX_LOCK(&stat_lock);
            miss_reply_count++;
            MUTEX_UNLOCK(&stat_lock);

            this->state = STATE_QUERY_ADDR;
        }


 assert(check_bkinfo());
 assert(bk->cup.num==bk->entry_use);

    }

    if(this->state == STATE_THINK) {
        Trace("sess# %d, condition leak\n", this->sid);
    }

get_last:

    if(bk) {
 assert(check_bkinfo());
 assert(bk->cup.num==bk->entry_use);
        UNLOCK_BK(bk);

        bk = NULL;
    }

#if 0
Trace("sess# %d, end of GET, state %d %s\n",
    this->sid, this->state, MSG_STATE_[this->state]);
#endif


        break;


    /* POST method */
    case M_POST:
Trace("POST method <%s>\n", this->loc);
        MUTEX_LOCK(&stat_lock);
        miss_reply_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->state = STATE_QUERY_ADDR;

        break;

    default:
        Error("sess# %d, ignore method %d '%s'",
            this->sid, this->state, MSG_STATE_[this->state]);

        MUTEX_LOCK(&stat_lock);
        miss_reply_count++;
        MUTEX_UNLOCK(&stat_lock);

        this->reason = R_UNSUPPORT_METHOD;
        this->state = STATE_MESSAGE;

        break;
    }

#if 0
    MUTEX_LOCK(&stat_lock);
    process_request_count++;
    MUTEX_UNLOCK(&stat_lock);
#endif

    return 0;
}


/*
#ifndef TRACE_SESS
#define TRACE_SESS
#endif
*/


int
DoCheckEndFriendSess(sess_t *this)
{
    struct timeval xbefore, xafter;
    int chk;
    bucket_t *bk;
    skcache_entry_t *stored_pos;
    int check_already_have;

    skcache_entry_t tmp;
    int idle;

#ifdef /*TRACE_SESS*/ 0
    Trace("DoCheckEndFriendSess: sess# %d, entry 0x%x, '%s' [%d:%d:%d:%d] %s\n",
        this->sid,
        this->cache_entry,
        this->loc,
        this->fhash, this->shash, this->thash, this->ffhash,
        MSG_STATE_[this->state]);
#endif

    bk = &skcache_bucket_base[this->fhash];

    LOCK_BK(bk);


    stored_pos = this->cache_entry;

    /* error */
    if(stored_pos==NULL) {
        Error("sess# %d, not found such entry", this->sid);

        goto do_retrieve;
    }

#ifdef TRACE_SESS
        Trace("stored %p ref %d hash %4d %4d %4d %4d %s\n",
            stored_pos, stored_pos->ref,
            stored_pos->h1, stored_pos->h2,
            stored_pos->h3, stored_pos->h4, stored_pos->name);
#endif

    if(stored_pos->h4!=this->ffhash||stored_pos->h3!=this->thash
        ||stored_pos->h2!=this->shash||stored_pos->h1!=this->fhash||
            strcmp(stored_pos->name, this->loc)!=0) {

        Trace("checkfriend: sess# %d, entry was changed. maybe expired\n",
            this->sid);

#ifdef TRACE_SESS
        Trace("stored %p %4d %4d %4d %4d %s\n",
            stored_pos,
            stored_pos->h1, stored_pos->h2,
            stored_pos->h3, stored_pos->h4, stored_pos->name);
#endif

        goto do_retrieve;
    }

#if 0
 assert(CHKTIME(stored_pos->hot));
    stored_pos->hot = time(NULL);
#endif

    if(stored_pos->owner==this) {
#ifdef TRACE_SESS
        Trace("sess# %d, the line is mine. *strange*\n", this->sid);
#endif
        goto do_retrieve;
    }
    else {
        if(stored_pos->cracked) {
#ifdef TRACE_SESS
            Trace("sess# %d, line was cracked\n", this->sid);
#endif
            goto do_retrieve;
        }
        else
#if 0
        if(stored_pos->ref==1 && stored_pos->owner==NULL)
#endif
        if(stored_pos->owner==NULL)
        {
#ifdef TRACE_SESS
            Trace("sess# %d, line is free\n", this->sid);
#endif
            stored_pos->owner = this;

            goto do_send;
        }
        else {
            idle = this->last_action - stored_pos->hot;
            if(idle < 0 && idle > 3*WFRIEND_TIMEOUT/4) {
#ifdef TRACE_SESS
                Trace("sess# %d, give up wait (%d sec)\n",
                    this->sid, idle);
#endif
                goto do_retrieve;
            }
            else {
#ifdef TRACE_SESS
                Trace("sess# %d, oid# %d is still hold by sess %p\n",
                    this->sid, stored_pos->oid, stored_pos->owner);
#endif
                /* not hit and not miss */
                goto still_wait;
            }
        }
    }

#ifdef TRACE_SESS
    Trace("reach to end\n");
#endif
    goto skip;

do_retrieve:
#ifdef TRACE_SESS
    Trace("do retrieve\n");
#endif
    if(stored_pos) {
#if 1
        stored_pos = NULL;
#endif
#if 1
        this->cache_entry = NULL;
#endif
    }

    this->cachable = SESS_UNCACHABLE;

    MUTEX_LOCK(&stat_lock);
    miss_reply_count++;
    MUTEX_UNLOCK(&stat_lock);

    this->state = STATE_QUERY_ADDR;
    MUTEX_LOCK(&stat_lock);
    nfriend--;
    MUTEX_UNLOCK(&stat_lock);

    goto skip;

do_send:
#ifdef TRACE_SESS
    Trace("do send\n");
#endif
    MUTEX_LOCK(&stat_lock);
    hit_reply_count++;
    MUTEX_UNLOCK(&stat_lock);

    this->state = STATE_SEND_CACHE;
    MUTEX_LOCK(&stat_lock);
    nfriend--;
    MUTEX_UNLOCK(&stat_lock);

    goto skip;

still_wait:
    /* not hit and not miss */
#ifdef TRACE_SESS
    Trace("still wait\n");
#endif

#if 0
    this->state = STATE_WAIT_FRIEND;
    MUTEX_LOCK(&stat_lock);
    nfriend++;
    MUTEX_UNLOCK(&stat_lock);
#endif

    goto skip;


skip:

#ifdef /*TRACE_SESS*/ 0
    if(this->state != STATE_WAIT_FRIEND) {
        Trace("DoCheckEndFriendSess: sess# %d, change to %s\n",
            this->sid, MSG_STATE_[this->state]);
    }
#endif
#if /*1*/ 0
    if(this->state != STATE_WAIT_FRIEND) {
        Trace("DoCheckEndFriendSess: sess# %d, change to %s\n",
            this->sid, MSG_STATE_[this->state]);
    }
#endif

    if(bk) {

 assert(bk->cup.num==bk->entry_use);
        UNLOCK_BK(bk);
        bk = NULL;
    }

out:

    return 0;
}


/*
#ifdef TRACE_SESS
#undef TRACE_SESS
#endif
*/


int
DoFindServer(sess_t *this)
{
    int chk;
    int ret;


#if 0
    Trace("DoFindServer: URL '%s'\n\t host '%s', port %d, path '%s'\n",
        this->loc, this->servername, this->serverport, this->serverpath);
#endif
    if(this->servername==NULL || this->servername[0]=='\0') {
        Error("sess# %d, empty hostname is specified", this->sid);
        return -1;
    }


#if 0
    hcache_list();
#endif

    chk = hcache_seek_addr_incache(this->servername, &(this->serveraddr));
    if(chk) {
        /* miss */
#if 0
        Trace("miss in hostname cache, '%s' (%d)\n",
            this->servername, chk);
        Trace("send query to resolver\n");
#endif
        chk = hcache_send_query_addr_direct(this->servername,
                &(this->serveraddr), &this->rfd);
#if 0
        Trace("    ret %d\n", chk);
#endif
        if(chk) {
            Error("fail hcache_send_query_addr_direct() (%d)\n", chk);
            this->reason = R_RESOLVER_DOWN;
            this->state = STATE_MESSAGE;
            goto out;
        }

        this->state = STATE_WAIT_ADDR;
        ret = 0;
    }
    else {
        /* hit */
#if 0
        Trace("hit in hostname cache, '%s'\n",
            this->servername);
#endif
        this->serveraddr.sin_family = AF_INET;

        this->state = STATE_CONN_SERV;
        ret = 0;
    }

out:
    return ret;
}


int
DoServerConnect(sess_t *this)
{
    int chk;

#if 0
    Trace("DoServerConnect: server '%s' [%s], port %d",
        this->servername, inet_ntoa(this->serveraddr.sin_addr),
        this->serverport);
#endif

#if 0
    this->last_action = time(NULL);
#endif

    /* asyncronus connect */

    /*    this->sfd = TCP_aconnect_addr(&this->serveraddr, this->serverport);*/

    if((this->dhr_hash == proxy_id) && (this->dhr_state != 'E')){
      if((this->dhr_state == 'A') || (this->dhr_state == 'B')){
	this->dhr_state = 'M';
	this->sfd = TCP_aconnect_addr
	  (&(cluster[this->dhr_hash_tmp].addr)
	   ,cluster[this->dhr_hash_tmp].port);
      }
    }
    else{
      if((this->dhr_state == 'A') || (this->dhr_state == 'B') ||
	 (this->dhr_state == 'M') || (this->dhr_state == 'S'))
	this->sfd = TCP_aconnect_addr
	  (&(cluster[this->dhr_hash].addr)
	   ,cluster[this->dhr_hash].port);
      if(this->dhr_state == 'E')
	this->sfd = TCP_aconnect_addr(&this->serveraddr, this->serverport);
    }

    if(this->sfd<0) {
        if(this->sfd == KOTETU_SOCK_SOCKETERROR) {
            this->reason = R_FAIL_SERVER_SOCKET;
#if 0
            exit(437);
#endif
        }
        else {
            Error("DoServerConnect: sess# %d, reason %d",
                this->sid, this->sfd);
            this->reason = R_FAIL_SERVER_CONNECT;
        }
        this->state = STATE_MESSAGE;
        return 0;
    }

    this->state = STATE_WAIT_CONN_SERV;

    return 0;
}

int
DoWaitServerConnect(sess_t *this)
{
    volatile int errno_backup;
    int err, len;
    int chk;


#if 0
    Trace("DoWaitServerConnect: server '%s' [%s], port %d\n",
        this->servername, inet_ntoa(this->serveraddr.sin_addr),
        this->serverport);
#endif

    chk = iswritable_via_poll(this->sfd, 1, 0);
    if(chk>0) {
        this->state = STATE_SEND_REQUEST;
    }

    return 0;
}

int
DoSendRequest(sess_t *this)
{
    char req[L_BUF];
    int len, rqlen;
    register char *p, *q, *r;
    int  nw;
    char line[STRING_SIZE];
    char *cpos;
    int c;


    /*
     * several server requires HOST field, independly protocol versions.
     * then, the program put HOST field.
     */
    if(this->method==M_POST) {
        sprintf(req, "POST %s HTTP/1.0\r\nHost: %s\r\n",
            this->serverpath, this->servername);
#if 0
    dump_stringchunk("orignal", this->reqbuf);
#endif
#if 0
    dump_stringsegment("orignal", this->reqbuf, this->reqlen);
    dump_stringchunk("first", req);
#endif
    }
    else {
      /*        sprintf(req, "GET %s HTTP/1.0\r\nHost: %s:%d\r\n",
		this->serverpath, this->servername, this->serverport);*/
      if(1)
          sprintf(req, "GET http://%s:%d%s HTTP/1.0\n\nXdhr: 2.0 %c\r\n",
		  this->servername, this->serverport,this->serverpath,
		  this->dhr_state);
      else
          sprintf(req, "GET %s HTTP/1.0\r\nHost: %s:%d\r\n",
		  this->serverpath, this->servername, this->serverport);
    }

#if 0
    dump_stringsegment("r.req.", this->reqbuf, this->reqlen);
#endif

    rqlen = HTTP_find_header_term(this->reqbuf, this->reqlen);
    if(rqlen>=L_REQBUF-1) {
        Error("too long requst (meybe rqlen is %d)", rqlen);
        dump_stringsegment("the request", this->reqbuf, this->reqlen);

        this->state = R_TOOLONG_REQ;
        this->state = STATE_MESSAGE;
        return -1;
    }

    p = this->reqbuf;
    r = this->reqbuf + rqlen;

    while(*p && *p!='\n') {
        p++;
    }
    p++;

    /* skip first line, because it was already copied */
    q = req;
    while(*q) {
        q++;
    }

    /* copy rest of header */
    cpos = p;
    while((cpos = sgets(p, cpos, line, STRING_SIZE))!=NULL) {
        if(line[0]=='\0') {
            break;
        }
        if((line[0]=='H' || line[0]=='h')
            && strncasecmp(line, "Host:", 5)==0) {
            continue;
        }
        if((line[0]=='P' || line[0]=='p')
            && strncasecmp(line, "Proxy-Connection:", 17)==0) {
            continue;
        }
        strcat(req, line);

        if(line[0]=='\r'||line[0]=='\n') {
            break;
        }
    }


    len = strlen(req);

#if 0
    dump_stringchunk("cooked request", req);
#endif

#if 1
    if(len>=L_REQBUF-1) {
        Error("too long requst (current %d, original %d)", len, this->reqlen);
        dump_stringchunk("the requst", req);

        this->state = R_TOOLONG_REQ;
        this->state = STATE_MESSAGE;
        return -1;
    }
#endif

    if(this->method==M_POST) {
        int c;

        c = this->reqlen - (cpos-&this->reqbuf[0]);

#if 0
        dump_stringchunk("cooked", req);
#endif
        if(c) {
#if 0
            dump_stringsegment("rest  ", cpos, c);
#endif
            p = cpos;
            q = req + len;
            while(c>0) {
                *q++ = *p++;
                c--;
                len++;
            }
#if 0
        dump_stringsegment("cooked", req, len);
#endif
        }
    }

#if 0
  if(this->mode == MODE_SELF ) {
    Trace("sess# %d, request final form, length %d\n", this->sid, len);
    dump_stringsegment("s.req.", req, len);
  }
#endif
#if 0
    Trace("sess# %d, request final form, mode %d, length %d\n",
        this->sid, this->mode, len);
    dump_stringsegment("s.req.", req, len);
#endif

    c = 0;
try_write:
Trace("passed try_write: errno %d\n",errno);
    c++;
    nw = write(this->sfd, req, len);
    if(nw<0) {
        if(errno==EINTR) {
            goto try_write;
        }

        if(errno==EPIPE) {
            this->reason = R_LOST_SERVER_CONNECT;
#if 0
            Error("sess# %d, server pipe (fd# %d) error (%d)",
                this->sid, this->sfd, errno);
#endif
        }
#if 0
        else
        if(errno==EAGAIN) {
        }
#endif
        else {
            this->reason = R_FAIL_SERVER_WRITE;

            Error("cannot write request to server sess %p fd #%d (%d)",
                this, this->sfd, errno);
        }

Trace("passed STATE_MESSAGE\n");
        this->state = STATE_MESSAGE;

        return -1;
    }
    if(nw==0) {
        this->state = STATE_CLOSE;
        return -1;
    }
#if 0
    Trace("write request %d byte(s)\n", nw);
#endif

    this->swrite += (off_t) nw;


#if 0
    this->last_action = time(NULL);
#endif


    if(this->method == M_GET) {
        this->state = STATE_RECV_RES;
    }
    else
    if(this->method == M_POST) {
        this->state = STATE_RELAY_DIRECT;
    }
    else {
    }

    return 0;
}



/*
 * for POST method
 */
int
DoRelayDirect(sess_t *this)
{
    int rest;
    int use_tmpbuf;
    char tmpbuf[L_BUF];
    int l;
    struct iovec iovec[2];
    int ciovec;
    int blen;
    int nr;
    int nw;
    int chk;
    int err;

    int dir;
    int cr, cw, sr, sw;

#if 0
    Trace("DoRelayDirect: self %#x, sess# %d, hash %d:%d\n",
        pthread_self(), this->sid, this->fhash, this->shash);
#endif



    dir = 0;
    cr = cw = sr = sw = (off_t)0L;

    /*
     * Since connection starts by client, client has high priority.
     */

client_side:
    if(this->cfd<0) {
        goto server_side;
    }
    chk = isreadable_via_poll(this->cfd, 0, 0);
    if(chk) {
        dir = 1; /* client */
    }
    else {
        goto server_side;
    }

try_c_read:
    nr = read(this->cfd, tmpbuf, L_BUF);
    if(nr<0) {
        if(errno==EINTR)
            goto try_c_read;

        this->reason = R_FAIL_CLIENT_READ;
        this->state = STATE_MESSAGE;
        goto out;
    }
    else
    if(nr==0) {
        DoHalfClientCloseSess(this);
        goto server_side;
    }


    cr += nr;

try_s_write:
    nw = write(this->sfd, tmpbuf, nr);
    if(nw<0) {
        if(errno==EINTR)
            goto try_s_write;

        this->reason = R_FAIL_SERVER_WRITE;
        this->state = STATE_MESSAGE;
        goto out;
    }

    sw += nw;


server_side:
    if(this->sfd<0) {
        goto disconn_check;
    }
    chk = isreadable_via_poll(this->sfd, 0, 0);
    if(chk) {
        dir += 2; /* server */
    }
    else {
        goto disconn_check;
    }

try_s_read:
    nr = read(this->sfd, tmpbuf, L_BUF);
    if(nr<0) {
        if(errno==EINTR)
            goto try_s_read;

        this->reason = R_FAIL_SERVER_READ;
        this->state = STATE_MESSAGE;
        goto out;
    }
    else
    if(nr==0) {
        DoHalfServerCloseSess(this);
        goto disconn_check;
    }

    sr += nr;

try_c_write:
    nw = write(this->cfd, tmpbuf, nr);
    if(nw<0) {
        if(errno==EINTR)
            goto try_c_write;

        this->reason = R_FAIL_CLIENT_WRITE;
        this->state = STATE_MESSAGE;
        goto out;
    }

    cw += nw;


disconn_check:

    if(this->cfd<0 || this->sfd<0) {
        this->state = STATE_CLOSE;
    }

out:
    if(cr>0) this->cread  += cr;
    if(cw>0) this->cwrite += cw;
    if(sr>0) this->sread  += sr;
    if(sw>0) this->swrite += sw;


#if 0
    Trace("sess# %d, dir %x, cfd,sfd=%d,%d r/w=c %d/%d, s %d/%d %s\n",
        this->sid, dir, this->cfd, this->sfd,
        cr, cw, sr, sw, MSG_STATE_[this->state]);
#endif

    return 0;
}



/*
 *
 */

int
DoServerReply(sess_t *this)
{
    int rest;
    int use_tmpbuf;
    char tmpbuf[L_BUF];
    int l;
    struct iovec iovec[2];
    int ciovec;
    int blen;
    int nr;
    int nw;
    int chk;
    int err;
    struct timeval xbefore, xafter;

    int hlen;
    int xvmajor, xvminor;
    time_t xdate, xlmd, xexpire;
    off_t xclen;
    int xrcode;
    int xcachable;
    int xreload;

    bucket_t *bk;
    skcache_entry_t *centry;
    int check_already_have;
    skcache_entry_t tmp;

    skcache_frag_t *flist;

    skcache_frag_t *last;
    skcache_frag_t *cur;
    off_t frag_sum;
    int u;

    int c;

    int res=-1;

/*
#ifdef TRACE_CACHE
    Trace("DoServerReply: self %#x, sess# %d, hash %d:%d\n",
        pthread_self(), this->sid, this->fhash, this->shash);
#endif
*/

    /***
     *** alloc buffer
     ***/
    bk = NULL;
    use_tmpbuf = 0;
    last = NULL;
    cur = NULL;
    blen = 0;
    centry = NULL;

    if(this->cachable==SESS_UNCACHABLE) {
#if 0
        Trace("sess# %d, uncachable\n", this->sid);
#endif
        res = 10;
        goto alloc_tmpbuf;
    }

    bk = &skcache_bucket_base[this->fhash];
    LOCK_BK(bk);

 assert(check_emptyentity(bk));

    centry = (skcache_entry_t*) this->cache_entry;
    if(centry==NULL) {
#if 0
        Trace("sess# %d, centry is NULL\n", this->sid);
#endif
        res = 20;
        goto alloc_tmpbuf;
    }
#ifdef CHECK_MAGIC
    if(centry->magic[0]!=MAGIC_0 || centry->magic[1]!=MAGIC_1) {
        Error("sr: ignore magic bk# %d, centry %p, sess# %d",
            bk->bid, centry, this->sid);
        centry = NULL;
        this->cache_entry = NULL;

        res = 30;
        goto alloc_tmpbuf;
    }
#endif
    if(centry->cracked) {
#if 0
        Trace("sess# %d, centry is cracked\n", this->sid);
#endif
        res = 40;
        goto alloc_tmpbuf;
    }

#if 0
    Trace("sess# %d, SR magic %02x %02x\n",
        this->sid, centry->magic[0], centry->magic[1]);
#endif


    /* seek last position of object frags in this object entry */
    /* retrieving object must be upon memory */

    last = NULL;
    frag_sum = 0L;
    u = 0;
    cur = centry->m_recorded;
    while(cur) {
        frag_sum += cur->len;
        last = cur;

        if(cur->next==cur) {
            Error("sess# %d, %dth frag %p is looped itself, %ld byte(s)",
                this->sid, u, cur, frag_sum);
            centry->cracked = 1;
            res = 50;
            goto alloc_tmpbuf;
        }

        cur = cur->next;

        u++;
        if(u>=nfrag_obj) {
#if 0
            Trace("sess# %d, too large, %d frags %ldB / max %d frags\n",
                this->sid, u, (long)frag_sum, nfrag_obj);
#endif
            res = 60;
            centry->cracked = 1;
            goto alloc_tmpbuf;
        }
    }

#if 0
    Trace("sr: sess# %d, u %d, cur %p, last %p\n",
        this->sid, u, cur, last);
#endif


    /* get new frag */
try_get_frag:
 assert(bk->cup.num==bk->entry_use);
    cur = skcache_pull_new_frag(bk);
 assert(bk->cup.num==bk->entry_use);
    if(cur==NULL) {
        Error("ServerReply: sess# %d, cannot get new frag on bk# %d, %p",
            this->sid, bk->bid, bk);

        centry->cracked = 1;
        res = 70;
        goto alloc_tmpbuf;
    }
    bk->frag_use++;

#if 0
Trace("  new frag from free-list %p <next %p>\n", cur, cur->next);
#endif

#if 0
    if(last) {
        if(last==cur) {
#if 0
            Error("PANIC !!! bk %p, last %p, cur %p, frag %d/%d %p",
                bk, last, cur,
                bk->frag_use, bk->frag_all, bk->frag_freelist);
#endif
            Error("PANIC !!! bk %p, last %p, cur %p, frag %d",
                bk, last, cur, bk->frag_use);

            skcache_dump_entry_frag("strange", centry);

            centry->cracked = 1;
            res = 80
            goto alloc_tmpbuf;
        }
    }
#endif

    if(last) {
        last->next = cur;
    }
    else {
        if(centry->m_recorded) {
            Trace(
"sr: assign current sess# %d, entry %p, oid# %d, m_recorded %p, cur %p,\n",
                this->sid, centry, centry->oid, centry->m_recorded, cur);
        }
        
        centry->m_recorded = cur;
    }

#if 0
    skcache_dump_entry_frag("alloc", centry);
#endif

    blen = 1;
    iovec[0].iov_base = cur->ent;
    iovec[0].iov_len  = cur->size;

    MUTEX_LOCK(&stat_lock);
    alloc_sess_buf_count++;
    MUTEX_UNLOCK(&stat_lock);

    goto alloc_done;



alloc_tmpbuf:
#if 0
    Trace("sess# %d, alloc tmpbuf\n", this->sid);
#endif

    if(centry) {
        int chk;

#if 0
        Trace("alloc_tmpbuf, sess# %d, centry %p, m_recorded %p\n",
            this->sid, centry, centry->m_recorded);
#endif
        if(centry->m_recorded && bk!=NULL) {
#if 0
            Trace("ATT: purge stored entry sess# %d, oid# %d, res %d\n",
                this->sid, centry->oid, res);
#endif
            centry->ref--;
            chk = skcache_entry_delete(bk, centry);
            if(chk) {
                Trace("  skcache_entry_delete() return %d\n", chk);
            }
        }
        centry = NULL;
        this->cache_entry = NULL;
    }

    blen = 1;
    iovec[0].iov_base = tmpbuf;
    iovec[0].iov_len  = sizeof(tmpbuf);

    MUTEX_LOCK(&stat_lock);
    alloc_tmp_buf_count++;
    MUTEX_UNLOCK(&stat_lock);

    goto alloc_done;


alloc_done:

    if(blen==1) {
        rest = iovec[0].iov_len;
    }
    else
    if(blen==2) {
        rest = iovec[0].iov_len + iovec[1].iov_len;
    }
    else {
        rest = -1;
    }

#if 0
    Trace("read:\n");
    Trace("  blen %d\n", blen);
    Trace("  iovec[0] base %p, len %d\n",
        iovec[0].iov_base, iovec[0].iov_len);
    if(blen>=2) {
        Trace("  iovec[1] base %p, len %d\n",
        iovec[1].iov_base, iovec[1].iov_len);
    }
    Trace("  rest %d\n", rest);
#endif

    /***
     *** read socket
     ***/
    MUTEX_LOCK(&stat_lock);
    read_reply_total++;
    MUTEX_UNLOCK(&stat_lock);

try_read:
    errno = 0;
    nr = readv(this->sfd, iovec, blen);
    err = errno;
#ifdef TRACE_CACHE
    if(use_tmpbuf) {
        Trace("  sess# %d, fd# %d, read %d/%d (temporaly area)\n",
            this->sid, this->sfd, nr, rest);
    }
    else {
        Trace("  sess# %d, fd# %d, read %d/%d (cache area)\n",
            this->sid, this->sfd, nr, rest);
    }
#endif
    if(nr<0) {
        if(err==EINTR) {
            goto try_read;
        }
        /* for OSF/1 */
        if(err==0) {
            goto try_read;
        }

        MUTEX_LOCK(&stat_lock);
        read_reply_fail++;
        MUTEX_UNLOCK(&stat_lock);

        switch(err) {
        case EAGAIN:
            goto out;
        case ECONNRESET:
            goto lost_connect;
        case EPIPE:
            Error("sr: sess# %d, pipe error fd# %d, sread %d",
                this->sid, this->sfd, this->sread);
            break;
        default:
            Error("sr: sess# %d, cannot read server reply fd# %d, sread %d (%d)",
                this->sid, this->sfd, this->sread, err);
        }

        goto fail_read;
    }

    MUTEX_LOCK(&stat_lock);
    read_reply_success++;
    MUTEX_UNLOCK(&stat_lock);

    if(nr==0) {
#if 0
        skcache_dump_entry_frag("before", centry);
#endif

        /* restore
         *  But if totallength is 0 (which means empty),
         *  We have to keep this one as mark of empty.
         */
        if(cur) {
            if((long)this->sread > 0L) {
#if 0
Trace("sr: flush sess# %d, sread %ld : cur %p centry %p: del\n",
                    this->sid, (long)this->sread,
                    cur, centry);
#endif

                skcache_push_idle_frag(bk, cur);
                cur = NULL;

                bk->frag_use--;

                blen = 0;

                if(last) {
                    last->next = NULL;
                }
                else {
                    if(centry && centry->totallength > 0) {
#if 1
Trace("sr: flush sess# %d, centry %p, oid# %d, m_recorded %p\n",
    this->sid, centry, centry->oid, centry->m_recorded);
#endif
                        centry->m_recorded = NULL;
                        centry->d_recorded = -1;
                    }
                }
            }
            else {
#if 0
Trace("sr: flush sess# %d, sread %ld : cur %p centry %p: keep\n",
                    this->sid, (long)this->sread,
                    cur, centry);
#endif
            }
        }
        else {
#if 0
Trace("sr: flush sess# %d, sread %ld : cur %p centry %p: skip\n",
                this->sid, (long)this->sread, 
                cur, centry);
#endif
        }

#if 0
        skcache_dump_entry_frag("after", centry);
        Trace("===\n");
#endif

        this->reason = -1;
        goto reach_end;
    }

#if 0
dump_stringsegment("rep", iovec[0].iov_base, MIN(nr, 128));
#endif

    this->sread += (off_t) nr;

    sim_stat_add_value(&len_perread_stat, (long)nr);

#if 0
    if(this->pos <= (off_t)0 && this->sver_major < 0)
#endif
    if(this->pos == (off_t)0 )
    {

#if 0
dump_stringsegment("first", iovec[0].iov_base, MIN(nr, 300));
#endif

        hlen = HTTP_find_header_term(iovec[0].iov_base,
                MIN(nr, iovec[0].iov_len));
#if 0
Trace("sess# %d, r header term %4d / %5d\n",
        this->sid, hlen, MIN(nr, iovec[0].iov_len));
#endif
        if(hlen>0) {
            if(hlen>=iovec[0].iov_len) {
Trace("sess# %d, r $ header term %4d / %5d\n",
        this->sid, hlen, MIN(nr, iovec[0].iov_len));
/*
                dump_stringsegment("broken",
                    iovec[0].iov_base, iovec[0].iov_len);
*/
            }

/*
            this->pos   = (off_t) hlen;
*/
            this->hlen  = hlen;
            xcachable   = this->cachable;
            xreload     = this->reload;

            chk = HTTP_parse_response_header(iovec[0].iov_base, hlen,
                    &xvmajor, &xvminor, &xrcode,
                    &xdate, &xlmd, &xexpire, &xclen,
                    &xcachable, &xreload);

            if(!chk) {
#if 0
Trace("%c(@", 0x1b);
Trace("%c$B%c(B", 0x1b, 0x1b);
#endif
#if 0
Trace("sess# %d, <hlen %d><chk %d> ver %d.%d, rcode %d, lmd %d, expire %d, clen %d\n",
    this->sid, hlen, chk,
    xvmajor, xvminor, xrcode, xlmd, xexpire, xclen);
#endif

                if(xrcode == 306){ 
                  skcache_entry_delete(bk, this->cache_entry);
/*		  skcache_push_idle_frag(bk, cur);
                  skcache_push_idle_entry(bk,this->cache_entry);
		  close(this->sfd);
		  this->sfd = -1;*/
		  DoHalfServerCloseSess(this);
		  this->dhr_state = 'E';
/*		  this->state = STATE_QUERY_ADDR;*/
		  this->state = STATE_THINK;
		  goto out;
		}


                this->sver_major    = xvmajor;
                this->sver_minor    = xvminor;
                this->rcode         = xrcode;
                this->lmd           = xlmd;
                this->expire        = xexpire;
                this->clen          = xclen;

#if 1
                if(centry) {
                    centry->lmd     = xlmd;
                    centry->expire  = xexpire;
                }
#endif


                if(xcachable==SESS_UNCACHABLE) {
#if 0
    Trace(" 74= sess# %d, %s\n", this->sid, this->loc);
    dump_stringsegment("req.", this->reqbuf, this->reqlen);
#endif
                    this->publish = SESS_PRIVATE;


                    if(centry) {
                        centry->cracked = 1;
                    }

                }
                else {
                    this->publish = SESS_PUBLIC;
                }

            }
            else {
                this->sver_major    = 0;
                this->sver_minor    = 0;
                this->rcode         = -1;
                this->lmd           = -1;
                this->expire        = -1;
                this->clen          = -1;
                this->publish       = SESS_PRIVATE;
            }

            this->pos += (off_t) this->sread;
        }
        else {
#if 0
Trace("sess# %d, hlen %d, nr %d, buf-len %d\n",
        this->sid, hlen, nr, iovec[0].iov_len);

dump_stringsegment("req.", this->reqbuf, this->reqlen);
#endif
        }

    }
    else {
        this->pos += (off_t) nr;
    }

#if 0
    if(this->rcode<0) {
        Trace("sess# %d, rcode %d\n", this->sid, this->rcode);
    }
#endif


    if(this->len<0) {
        this->len = (off_t) nr;
    }
    else {
        this->len += (off_t) nr;
    }

    if(centry) {
        if(centry->totallength<=0)
            centry->totallength = nr;
        else
            centry->totallength += nr;

        if(blen==1) {
            if(cur) {
                cur->len += nr;
            }
            else {
                Error("no cur frag to record length");
            }
        }
        else {
            Error("ignore blen %d", blen);
        }
#if 0
        else
        if(blen==2) {
            if(nr <= iovec[0].iov_len) {
                tlpos->cntl.length += nr;
            }
            else {
                tlpos->cntl.length += iovec[0].iov_len;
                nlpos->cntl.length = (nr - iovec[0].iov_len);
            }
        }
#endif
    }

#if 0
    else {
        Trace("does not have centry to save, sess# %d\n", this->sid);
    }
#endif

#if 0
    this->last_action = time(NULL);
#endif
    if(centry) {
        centry->hot = this->last_action;
 assert(CHKTIME(centry->hot));
    }

    if(this->cfd>=0) {
        if(blen==2) {
            if(nr<=iovec[0].iov_len) {
                blen = 1;
                iovec[0].iov_len = nr;
            }
            else {
                if(nr<=iovec[0].iov_len + iovec[1].iov_len) {
                    iovec[1].iov_len = nr - iovec[0].iov_len;
                }
                else {
        Error("length of read <%d> is large than buffer <%d(%d+%d)>\n",
            nr, iovec[0].iov_len + iovec[1].iov_len);
                }
            }
        }
        else
        if(blen==1) {
            if(nr<iovec[0].iov_len) {
                iovec[0].iov_len = nr;
            }
        }
#if 0
Trace("write:\n");
Trace("  blen %d\n", blen);
Trace("  iovec[0] base %p, len %d\n", iovec[0].iov_base, iovec[0].iov_len);
if(blen>=2)
Trace("  iovec[1] base %p, len %d\n", iovec[1].iov_base, iovec[1].iov_len);
#endif

try_write:
        nw = writev(this->cfd, iovec, blen);

        if(nw<0) {
            if(errno==EINTR) {
                goto try_write;
            }
            if(errno==EPIPE) {
                this->reason = -2;
                DoHalfClientCloseSess(this);
                goto reach_end;
            }

            Error("sr: sess# %d, cannot write to client fd# %d (%d)",
                this->sid, this->cfd, errno);
#if 0
            this->state = STATE_CLOSE;
#endif
            this->reason = R_FAIL_CLIENT_WRITE;
            DoHalfClientCloseSess(this);

            goto fail_write;
        }
        if(nw<nr) {
            Error("sr: sess# %d, wrote size (%d) is smaller than read size (%d)",
                this->sid, nw, nr);
            goto fail_write;
        }
        if(nw>nr) {
            Error("sr: sess# %d, wrote size (%d) is larger than read size (%d)",
                this->sid, nw, nr);
        }
        if(nw==0) {
            Error("sr: sess# %d, wrote size (%d) is zero blen %d",
                this->sid, nw, blen);
        }

/*
        this->pos += (off_t) nw;
*/
        this->cwrite += (off_t) nw;

#if 0
Trace("sess# %d, nw %ld, cwrite %ld\n", (long)nw, (long)this->cwrite);
#endif

    }


#if 0
    if(this->pos != this->len) {
    Trace("sess# %d, pos %ld, len %ld\n",
        this->sid, (long)this->pos, (long)this->len);
    }
#endif

    goto out;

/*
 * error recovery
 */

fail_memory:
    Error("sr: fail memory arrengment or memory trouble, sess# %d",
        this->sid);
    this->reason = R_LOST_BLOCK;

    goto fail;

lost_connect:
    Error("sr: lost conection establishment to server with some reason, sess# %d",
        this->sid);
    this->reason = R_LOST_SERVER_CONNECT;

    goto fail;

fail_connect:
    Error("sr: fail conection establishment to server with some reason, sess# %d",
        this->sid);
    this->reason = R_FAIL_SERVER_CONNECT;

    goto fail;

fail_read:
    Error("sr: fail reading from server with some reason, sess# %d",
        this->sid);
    this->reason = R_FAIL_SERVER_READ;

    goto fail;

fail_write:
    Error("sr: fail writing to client with some reason, sess# %d",
        this->sid);
    this->reason = R_FAIL_CLIENT_WRITE;

    goto fail;

fail:
    this->state = STATE_MESSAGE;

    goto out;




reach_end:
    this->state = STATE_CLOSE;
#ifdef TRACE_CACHE
    Trace("Ok,sess# %d [%d:%d]                 read %12ld / %4ldK / %4ldM\n",
        this->sid, this->fhash, this->shash,
        (long)this->len,
        (long)dK(this->len), (long)dM(this->len) );
#endif

    sim_stat_add_value(&object_size_stat, (long)(this->len));

    if(this->len==0) {
        Log("; sess# %d, empty ?", this->sid);
    }

#if 0
    Trace("sr: sess# %d, %d %d diff %d, cl %ld, hl %ld, sr %ld, cw %ld, pos %ld\n",
        this->sid,
        (this->sread==this->cwrite && this->cwrite==this->pos),
        (this->clen+this->hlen + 1 == this->sread),
        (long)(this->clen + this->hlen + 1 - this->sread),
        (long)this->clen,
        (long)this->hlen,
        (long)this->sread, (long)this->cwrite, (long)this->pos);
#endif

    if(centry) {
        /* discard cached object */
        if(this->rcode == 304) {
            centry->cracked = 1;
        }

#if 0
        Trace("sess# %d, length server said/recoreded %12ld/%12ld\n",
            this->sid, (long)this->clen, (long)centry->totallength);
#endif

        centry->owner = NULL;
    }

    goto out;

out:
    if(bk) {

 assert(bk->cup.num==bk->entry_use);
        UNLOCK_BK(bk);
        bk = NULL;
    }
    else {
#if 0
 Trace("bk is empty\n");
#endif
    }

    if(centry) {
#if 0
        centry->hot = xafter.tv_sec;
#endif
        centry->hot = this->last_action;
 assert(CHKTIME(centry->hot));
    }
    else {
#if 0
 Trace("centry is empty\n");
#endif
    }

#ifdef TRACE_CACHE
Trace(" @\n\n");
#endif

    return 0;
}



int
DoSendCache(sess_t *this)
{
    int chk;
    int sum;
    int nw;
    int l;
    struct iovec iov[HARD_NFRAG_NWRITE];
    int ciov;
    time_t now;
    int hlen;
    int xvmajor, xvminor, xrcode;
    time_t xdate, xlmd, xexpire;
    off_t xclen;
    int xcachable, xreload;
    bucket_t *bk;
    skcache_entry_t *centry;
    skcache_frag_t *cur;
    skcache_frag_t *first;


    bk  = &skcache_bucket_base[this->fhash];
    LOCK_BK(bk);

    now = time(NULL);

 assert(check_emptyentity(bk));

    centry = this->cache_entry;
    if(centry==NULL) {
        goto tobe_miss;
    }
#if 0
    Trace("stored %p oid# %d, ref %d, [%d:%d:%d:%d] %s\n",
        centry, centry->oid, centry->ref,
        centry->h1, centry->h2, centry->h3, centry->h4, centry->name);
#endif

    if(centry->ref<0) {
        Trace("sc: hit sess# %d [%d:%d:%d:%d], it was dead. maybe expired\n",
            this->sid, this->fhash, this->shash, this->thash, this->ffhash);
        goto tobe_miss;
    }

#if 0
    Trace("comp centry %x vs this\n\
%d/%d %d/%d %d/%d %d/%d\n  %s\n  %s\n",
    centry,
    centry->h1, this->fhash,
    centry->h2, this->shash,
    centry->h3, this->thash,
    centry->h4, this->ffhash,
    centry->name, this->loc);
#endif

    if(centry->h4!=this->ffhash||centry->h3!=this->thash
        ||centry->h2!=this->shash||centry->h1!=this->fhash||
            strcmp(centry->name, this->loc)!=0) {

    Trace("sc: hit sess# %d [%d:%d:%d:%d], entry was changed. maybe expired\n",
            this->sid, this->fhash, this->shash, this->thash, this->ffhash);

#if 0
        Trace("comp centry %x vs this\n\
%d/%d %d/%d %d/%d %d/%d\n  %s\n  %s\n",
    centry,
    centry->h1, this->fhash,
    centry->h2, this->shash,
    centry->h3, this->thash,
    centry->h4, this->ffhash,
    centry->name, this->loc);
#endif

        goto tobe_miss;
    }


 assert(bk->cup.num==bk->entry_use);

    if(centry->m_recorded == NULL) {
        if(centry->d_recorded>=0) {
#if 1
            Trace("DoSendCache: hit sess# %d that object is stored in disk\n",
                this->sid);
#endif

            chk = skcache_entry_swapin(bk, centry);
#if 1
            Trace("DoSendCache: sess# %d, that object is stored in disk, %d\n",
                this->sid, chk);
#endif
            if(chk) {
                Trace("sess# %d, fail to swapin oid# %d from disk <%d>\n",
                    this->sid, centry->d_recorded, chk);
                first = NULL;
            }
            else {
                first = centry->m_recorded;
#if 0
                centry->oid = this->sid;
#endif
            }
        }
        else {
            first = NULL;
        }
    }
    else {
        first = centry->m_recorded;
    }

 assert(bk->cup.num==bk->entry_use);

    /* not found */
    if(first==NULL) {
        goto tobe_miss;
    }

#if 0
    Trace("DoSendCache: sess# %d, len %d, pos %d\n",
        this->sid, this->len, this->pos);
#endif

#if 0
    Trace("  object# %d <%p>, '%s'\n\
  create time %d, hot %d, lmd %d, expire %d\n\
  totallength %d\n",
        centry->oid, centry, centry->name,
        centry->issue_time, centry->hot, centry->lmd, centry->expire,
        centry->totallength);
#endif

    if(centry->oid<0) {
        Error("sendcache: sess# %d, stored oid is negative oid# %d",
            this->sid, centry->oid);

#if 1
    Trace("  object# %d <%p>, '%s'\n\
  create time %d, hot %d, lmd %d, expire %d\n\
  totallength %d\n",
        centry->oid, centry, centry->name,
        centry->issue_time, centry->hot, centry->lmd, centry->expire,
        centry->totallength);
#endif

        goto tobe_miss;
    }

    /*
     * length check
     */

    if(centry->totallength<=0) {
#if 0
        Trace("sess# %d, length<=0 (length %d, total %d, clen %ld)\n",
            this->sid,
            hlpos->cntl.length, hlpos->cntl.totallength, (long)this->clen);
#endif
        goto tobe_miss;
    }

    this->oid = centry->oid;

  if((off_t)this->pos==(off_t)0L) {

#if 0
    Trace("sc: sess# %d, first->ent %p, len %d/%d, this->pos %ld\n",
        this->sid, first->ent, first->len, first->size, (long)this->pos);
#endif

    hlen = HTTP_find_header_term( (char*)(first->ent + this->pos), first->len);
    if(hlen<=0) {
#if 1
        Trace("sc: sess# %d, oid# %d, not found header %d/%d\n",
            this->sid, this->oid, hlen, first->len);
#endif

#if 0
        dump_stringsegment("sc:header", (char*)(first->ent + this->pos), first->len);
#endif

        centry->cracked = 1;
        goto tobe_miss;
    }

    chk = HTTP_parse_response_header(
            (char*)(first->ent + this->pos), hlen,
            &xvmajor, &xvminor, &xrcode,
            &xdate, &xlmd, &xexpire, &xclen,
            &xcachable, &xreload);

    if(chk) {
#if 1
        Trace("sc: sess# %d, ignore header\n", this->sid);
#endif
#if 0
        dump_stringsegment("sc:header", (char*)(first->ent + this->pos), hlen);
#endif

        goto tobe_miss;
    }
    else {
        this->clen = xclen;
    }
#if 0
    Trace("sc: version %d.%d, rcode %d\n", xvmajor, xvminor, xrcode);
    Trace("sc: lmd %d, expire %d, clen %ld\n", xlmd, xexpire, (long)xclen);
#endif

    if(xrcode<0) {
#if 0
        Trace("sess# %d, rcode is ignore (%d)\n",
            this->sid, xrcode);
#endif

        goto tobe_miss;
    }
    this->rcode = xrcode;

    if(this->clen>0) {
        if((centry->totallength-hlen-1)!=this->clen) {
 Error("sc: sess# %d, oid# %d, size miss match stored %ld header said %ld",
                this->sid, centry->oid,
                (long)centry->totallength-hlen-1, (long)this->clen);
#if 0
        Trace("sc: totallength %ld, hlen %ld, clen %ld\n",
            (long)centry->totallength, (long)hlen, (long)this->clen);
        dump_stringsegment("sc:header", (char*)(first->ent + this->pos), hlen);
#endif

            centry->cracked = 1;

            goto tobe_miss;
        }
        else {
        }
    }

    this->len = centry->totallength;


    /* cracked check */

    if(centry->cracked) {
        goto tobe_miss;
    }


    /* expire check */

    if(centry->expire>0 && this->last_action > centry->expire) {
        goto tobe_miss;
    }


    /*
     *
     */
    if(this->ims>0) {
#if 0
        Trace("sc: sess# %d, ims %d, lmd %d, expire %d\n",
                this->sid, this->ims, centry->lmd, centry->expire);
#endif
    }


    /*
     *
     */
#if 0
    centry->count++;
    this->count = centry->count;
    centry->hot = this->last_action;
 assert(CHKTIME(centry->hot));

    this->ishit = SESS_HIT;
    this->oid = centry->oid;
#endif

    centry->hot = this->last_action;
 assert(CHKTIME(centry->hot));

    if(bk) {
    }

 }
 else {
#if 0
    Trace("sc: sess# %d, sent %ld/%ld\n",
        (long)this->pos, (long)centry->totallength);
#endif
 }


    if(this->pos>=centry->totallength) {
#if 1
        Trace("sc: sess# %d, all content of object was send pos/len=%d/%d\n",
            this->sid, this->pos, this->len);
#endif
        goto done;
    }


    /*
     * skip current point to last sending
     */
    if(this->pos>(off_t)0)
    {
        off_t sum;
        int a;

        sum = (off_t)0L;
        cur = first;
        a = 0;

        do {
            sum += cur->len;

            if(cur==cur->next) {
                Error("sc; frag looped, broken data was send");
                centry->cracked = 1;

                goto done;
            }

            cur = cur->next;
            a++;
        } while(cur && sum<this->pos);

        if(cur==NULL) {
            Error("sess #d, maybe reach to end", this->sid);
            goto done;
        }
    }
    else {
        cur = first;
    }


    if(bk) {
    }

    /*
     *
     */

    sum = 0;
/*
    cur = first;
*/

    do
    {
        l = 0;
        do {
            iov[l].iov_base = cur->ent;
            iov[l].iov_len  = cur->len;
            if(cur->len>0) {
                sum += cur->len;
            }

            l++;

            cur = cur->next;
        } while(cur && l < nfrag_nwrite && cur->len>0 );
        ciov = l;

        if(this->cfd<0) {
            break;
        }

try_write:
        nw = writev(this->cfd, iov, ciov);
        if(nw<0) {
            if(errno==EINTR) {
                goto try_write;
            }
            if(errno==EPIPE) {
                DoHalfClientCloseSess(this);
                goto done;
            }
            Error("sess# %d, cannot write to client fd# %d, ciov %d, (%d)",
            this->sid, this->cfd, ciov, errno);
            this->reason = R_FAIL_CLIENT_WRITE;
            this->state = STATE_CLOSE;

            goto out;
        }

        this->cwrite += (off_t) nw;
        this->pos += nw;

    } while(sum<nfrag_nwrite && cur!=NULL);

    this->last_action = time(NULL);

    if(this->pos>=centry->totallength) {
        goto done;
    }

still:
    /* do nothing */
    goto out;

done:
    this->state = STATE_CLOSE;
    goto out;


tobe_miss:

    /* when enter this routine, the program increment hit counter.
     * if miss, the program has to decrement that and increment miss counter.
     */
    MUTEX_LOCK(&stat_lock);
    miss_reply_count++;
    hit_reply_count--;
    skip_cache_count++;
    MUTEX_UNLOCK(&stat_lock);


#if 0
    if(centry && bk) {
        Trace("sc:miss: to be miss sess# %d, oid# %d, cwrite %ld, pos %ld\n",
            this->sid, centry->oid, (long)this->cwrite, (long)this->pos);
        centry->ref--;
        chk = skcache_entry_delete(bk, centry);
        if(chk) {
            Trace("sc:miss: skcache_entry_delete() return %d\n", chk);
        }
        centry = NULL;
    }
    this->cache_entry = NULL;
    this->state = STATE_QUERY_ADDR;
#endif

    if(centry) {
        Trace(
		  "sc:miss: to be miss sess# %d, hf %d, oid# %d, cwrite %ld, pos %ld\n",
          this->sid, this->hf, centry->oid, (long)this->cwrite, (long)this->pos);
        fflush(stderr);
        skcache_entry_show(centry, stdout);
        fflush(stdout);

        if((long)this->cwrite>0L) {
            Trace("sc:miss: during sending. rest of buffer is lost\n");
        }
        else {
        }
    }
	else {
        Trace(
		  "sc:miss: to be miss sess# %d, hf %d, oid# ?, cwrite %ld, pos %ld\n",
          this->sid, this->hf, (long)this->cwrite, (long)this->pos);
	}
    this->state = STATE_QUERY_ADDR;

    goto out;

out:
    if(bk) {
 assert(bk->cup.num==bk->entry_use);
        UNLOCK_BK(bk);
    }

    return 0;
}





/* support only IPv4 */
static
int
store_scrambled_name(char *dst, struct sockaddr_in *addr, int sc_key)
{
    register char *p, *q;
    register int i;
    register int c;

    q = (char*) &addr->sin_addr;

    c = 0xc634d3a5;
    c |= sc_key;
    for(i=3;i>=0;i--) {
        p = &q[i];
        c += *p;
        c = c << 3 + 0x01;

#if 0
        Trace(" #%d %3u %2x -> %12u %12x\n",
            i, (unsigned int)*p, (unsigned int)*p,
            (unsigned int)c, (unsigned int)c);
#endif
    }

    sprintf(dst, "c%u", c);

    return 0;
}

static
int
store_direct_name(char *dst, struct sockaddr_in *addr)
{
    register char *p, *q;
    register int i;
    register int c;

    p = (char*) &addr->sin_addr;
    q = dst;

    for(i=0;i<4;i++) {
        if(*p>100)
            *q++ = '0' + (*p / 100);
        if(*p>10)
            *q++ = '0' + ((*p % 100)/10);
        *q++ = '0' + (*p % 10);
        *q++ = '.';
        *q = '\0';

#if 0
        Trace(" #%d %3u %2x -> '%s'\n",
            i, (unsigned int)*p, (unsigned int)*p, dst);
#endif

        p++;
    }

    return 0;
}


static
int client_scramble_key;

void*
Listener(void *called_option)
{
    int chk;
    struct sockaddr_in from;
    int len;
    int cfd;
    int find;
    int p, c;
    int op;
    int listen_fd;
    sess_t *sessp;
    int t_req, t_rjq;
    int t_nhot;
    int err;
    int try_c;
    struct timeval tmp_accept_time;
    char tmp_clientname[L_HOST];
    unsigned char *caddr;
    off_t nr, nw;


    listen_fd = *(int*)called_option;
    len = sizeof(from);

    while(1) {
try_accept:
#ifdef DO_ACCEPT_LOCK
        chk = MUTEX_LOCK(&accept_lock);
        if(chk) {
            Error("Listener: fail lock (%d)", chk);
            if(chk==EBUSY) {
                Trace(" ---  busy\n");
                goto try_accept;
            }

            goto try_accept;
        }
#endif

        err = errno = -1;
        cfd = accept(listen_fd, (struct sockaddr*)&from, &len);
        err = errno;
        gettimeofday(&tmp_accept_time, NULL);

#if 0
        Trace("Listener<%p> accept fd# %d (errno %d)\n",
            pthread_self(), cfd, err);
#endif

#ifdef DO_ACCEPT_LOCK

        try_c=0;
try_unlock:
        try_c++;

        chk = MUTEX_UNLOCK(&accept_lock);
        if(chk) {
            Error("Listener: fail unlock (%d) [%d]", chk, try_c);
            goto try_unlock;
        }
#endif

        if(cfd<0) {
            if(err==EINTR)
                goto try_accept;
#if 0
            if(err==ENFILE || err==EMFILE) {
                sleep(5);
                /*
                exit(143);
                */
            }
#endif

            Error("accept error (%d)", err);

            goto next;
        }


        MUTEX_LOCK(&stat_lock);
        receive_request_count++;
        request_sirial_count++;
        t_req = request_sirial_count;
        MUTEX_UNLOCK(&stat_lock);

#if 0
        Trace("th %d, conn. count %6d, cfd %3d\n",
            pthread_self(), t_req, cfd);
#endif


        /*
         * store peer (client or lower proxy) name
         */
#ifdef LOG_DIRECTCLIENTIPADDRESS
        store_direct_name(tmp_clientname, &from);
#else
        store_scrambled_name(tmp_clientname, &from, client_scramble_key);
#endif

#if 0
        Trace("peer <%s>\n", tmp_clientname);
#endif


        /*
         * check permition
         */
        caddr = (unsigned char*) &from.sin_addr.s_addr;
        if(!Acceptable(caddr)) {
            goto not_permitted;
        }


found_free:

        sessp = (sess_t*)sq_get_free();
        if(sessp==NULL) {
            Error("not enough memory for session");
            goto not_found_free;
        }


        /*
         * NOTE:
         *   # of hot session will be discrease.
         *   however, it is not increase. because increasing is only
         *   happen in this routines.
         */

        MUTEX_LOCK(&sess_lock);
        nhot++;
        MUTEX_UNLOCK(&sess_lock);



        chk = pthread_mutex_init(&(sessp->busy), NULL);

        sessp->sid    = t_req;
        sessp->psid   = -1;
        sessp->mode   = MODE_CLIENT;
        sessp->level  = 0;
        sessp->state  = STATE_ACCEPT;

        sessp->accept_time = tmp_accept_time;
        sessp->last_action = tmp_accept_time.tv_sec;

        sessp->reported = 0;
        sessp->reason = -1;

		sessp->hf	= 0;

        sessp->pos    =  0;
        sessp->len    = -1;
        sessp->rcode  = -1;
        sessp->count  = 0;
        sessp->lmd    = -1;
        sessp->expire = -1;

        sessp->cver_major = -1;
        sessp->cver_minor = -1;
        sessp->sver_major = -1;
        sessp->sver_minor = -1;

        sessp->cachable = SESS_CACHABLE;
        sessp->reload   = SESS_NORELOAD;
        sessp->ishit    = SESS_MISS;
        sessp->publish  = SESS_PUBLIC;

        sessp->cache_entry = NULL;

        sessp->oid    = -1;

        sessp->cfd    = cfd;
        sessp->sfd    = -1;
        sessp->rfd    = -1;
        sessp->rcount = 3;

        strcpy(sessp->clientname, tmp_clientname);

        memset(sessp->loc, 0, sizeof(sessp->loc));
        sessp->method = M_UNKNOW;

        memset(sessp->reqbuf, 0, sizeof(sessp->reqbuf));
        sessp->reqlen = 0;

        sessp->ims = -1;

#ifdef TRACE_SESS
        Trace("sess# %d, new, addr %p\n", sessp->sid, sessp);
#endif

        sessp->state  = STATE_RECV_REQ;


        chk = push_sess(think_queue, sessp);
        if(chk) {
            Error("Listener: sess# %d, fail push to think Q", sessp->sid);
            sessp->reason = R_MEMORY_OVERFLOW;

            DoSendMessage(sessp);
            DoCloseSess(sessp);
            DoReleaseSess(sessp);
            sessp = NULL;

            goto next;
        }

#if 0
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

        goto next;



not_found_free:
#if 0
        Error("not found idle session cell");
#endif
        MUTEX_LOCK(&stat_lock);
        reject_request_count++;
        t_rjq = reject_request_count;
        MUTEX_UNLOCK(&stat_lock);
#if 1
        if(t_rjq%100==0) {
            LogLine(": ; session list overflow");
        }
#endif


        nw = HTTP_SendIgnoreMessage(cfd, 501,
            "service denided", "please wait few minutus");
        shutdown(cfd, 2);
        close(cfd);
        cfd = -1;

        MUTEX_LOCK(&stat_lock);
        intra_outgoing += (off_t) nw;
        MUTEX_UNLOCK(&stat_lock);

        goto next;



not_permitted:
#if 0
        Error("not permitted host");
#endif
        MUTEX_LOCK(&stat_lock);
        reject_request_count++;
        t_rjq = reject_request_count;
        MUTEX_UNLOCK(&stat_lock);
#if 1
        LogLine(": ; request from not permitted host <%s>");
#endif

        nw = HTTP_SendIgnoreMessage(cfd, 403,
            "access denided", "you are not permitted this server");
        shutdown(cfd, 2);
        close(cfd);
        cfd = -1;

        MUTEX_LOCK(&stat_lock);
        intra_outgoing += (off_t) nw;
        MUTEX_UNLOCK(&stat_lock);

        goto next;



next:
        chk = 0;

    }

}



int
DoParseAnchor(sess_t *this)
{
    int chk;
    int rest;
    int nw;
    int l;

    bucket_t *bk;
    skcache_entry_t *centry;
    skcache_frag_t *first, *cur;

    int t_req=1234; /* dummy */

    int nc;
    int ntag;
    int ihave;

    int howprefetch;

    tags_t  *tags;
    char    **urllist;
    int     nofref;

    sess_t *psessp;
    char **aurl;
    char nurl[L_URL];

    int fh, sh, th, ff;

#ifdef TRACE_THREADS
    Trace("DoParseAnchor: sid %d, [%d:%d]\n",
        this->sid, this->fhash, this->shash);
#endif



    tags = NULL;

    if(this->cachable==SESS_UNCACHABLE) {
#ifdef TRACE_PREFETCH
        Trace("DoParseAnchor: sess# %d is not cachable\n", this->sid);
#endif
        goto out_nothing;
    }
#ifdef TRACE_PREFETCH
    Trace("DoParseAnchor: sess# %d, level %d, URL '%s'\n",
        this->sid, this->level, this->loc);
#endif

    /* allow level 0 or 1 */
    if(this->level<0 || this->level>2) {
#if 0
        Trace("sess# %d %p, strange level %d\n",
            this->sid, this, this->level);
#endif
        goto out;
    }

    /*
     *
     */
    bk = &skcache_bucket_base[this->fhash];
    LOCK_BK(bk);
#if 0
    /* XXX */
    if(chk) {
        Trace("DoParseAnchor: fail lock block #%d [%d:%d:%d]\n",
            this->fhash, this->fhash, this->shash, this->thash);
        goto out_pending;
    }
#endif

    centry = (skcache_entry_t*) this->cache_entry;

    /* not found object */
    if(centry==NULL) {
#if 0
        Trace("sess# %d, no entry in block# %d [%d:%d:%d]\n",
            this->sid, this->fhash, this->fhash, this->shash, this->thash);
#endif
#if 0
        cache_showblockaddr(this->fhash, 0);
#endif

 assert(bk->cup.num==bk->entry_use);
        UNLOCK_BK(bk);
        bk = NULL;
        goto not_found;
    }

    first = centry->m_recorded;
/* XXX */
#if 0
    if(centry->d_recorded>=0) {
    }
#endif
    if(first==NULL) {
        Trace("sess# %d, no frag in block# %d [%d:%d]\n",
            this->sid, this->fhash, this->fhash, this->shash);
#if 0
        cache_showblockaddr(this->fhash, 0);
#endif

 assert(bk->cup.num==bk->entry_use);
        UNLOCK_BK(bk);
        bk = NULL;
        goto not_found;
    }


    /***
     ***
     ***/

    tags = HTML_NewTags();

    if(this->level==0) {
#ifdef TRACE_PREFETCH
        Trace("parse refered: '%s' %db\n",
            this->loc, first->length);
#endif
        chk = MUTEX_LOCK(&config_lock);
        howprefetch = num_prefetch_referedpages;
        chk = MUTEX_UNLOCK(&config_lock);

        ntag = HTML_ParseTags_FromMemory(tags, PARSE_REFERED,
                first->ent, first->len);
#ifdef TRACE_PREFETCH
        Trace("parse refered: #%d %ldb %d\n",
            this->sid, first->length, ntag);
#endif
    }
    else
    if(this->level==1) {
#ifdef TRACE_PREFETCH
        Trace("parse included: '%s' %db\n",
            this->loc, first->length);
#endif
        chk = MUTEX_LOCK(&config_lock);
        howprefetch = num_prefetch_includedobjs;
        chk = MUTEX_UNLOCK(&config_lock);

        ntag = HTML_ParseTags_FromMemory(tags, PARSE_INCLUDED,
                first->ent, first->len);

#ifdef TRACE_PREFETCH
        Trace("parse included: #%d %db %d\n",
            this->sid, first->length, ntag);
#endif
    }


#if 0
    dump_stringchunk("entity", (char*)&line_pos->body);
    Trace("there are %d tag(s).\n", ntag);
#endif

 assert(bk->cup.num==bk->entry_use);
    UNLOCK_BK(bk);
    bk = NULL;
    centry = NULL;
    first = NULL;




    if(ntag<=0) {
        FREE(tags);
        tags = NULL;
        goto out;
    }

    if(tags==NULL) {
        goto out;
    }
    else {
        if(tags->count==0) {
            FREE(tags);
            tags = NULL;
            goto out;
        }
    }

    urllist = HTML_ConvURLList(tags, &nofref);
    HTML_DestroyTags_withoutvalue(tags);
    tags = NULL;



    /*
     *
     */

#if 0
    /*
    HTML_ShowIndependURLList(this->loc, urllist);
    */
    HTML_ShowURLList(urllist);
#endif

#ifdef TRACE_PREFETCH
    if(urllist) {
        Trace("DoParseAnchor<%d>: decide prefetch targets about\n  '%s'\n",
            pthread_self(), this->loc);
    }
    else {
        Trace("DoParseAnchor<%d>: prefetch target list is empty\n");
    }
#endif

    nc = 0;
    aurl = urllist;
    while(*aurl!=NULL && nc<howprefetch) {
#if 0
        Trace("  newone '%s'\n", *aurl);
#endif


        URL_JoinStr(nurl, L_URL, this->loc, *aurl);
        if(nurl[0]=='\0') {
            goto next;
        }
#ifdef TRACE_PREFETCH
        Trace("  full URL '%s'\n", nurl);
#endif

        calc_hash(nurl, &fh, &sh, &th, &ff);
#if 0
Trace("pan: %03d:%03d:%03d:%03d %s\n",
        fh, sh, th, ff, this->loc);
#endif

#if 0
        Trace("  check object existance in cache\n");
#endif
#if 0
        ihave = cache_have(fh, sh, nurl, 1, 0);
#endif
        ihave = skcache_have(fh, sh, th, ff, nurl, 1, 0);

#if 0
        Trace("  ihave %d\n", ihave);
#endif
        if(ihave>0) {
            goto next;
        }
        /* URL is not empty and it is not hold in cache */
        else
        if(ihave==0) {
            psessp = (sess_t*)sq_get_free();
            if(psessp==NULL) {
                Error("not enough memory for new prefetching session");
                goto skip;
            }

            chk = pthread_mutex_init(&(psessp->busy), NULL);

            MUTEX_LOCK(&sess_lock);
            nhot++;
            MUTEX_UNLOCK(&sess_lock);


#if 1
            MUTEX_LOCK(&stat_lock);
            request_sirial_count++;
            t_req = request_sirial_count;

            issue_prefetch_count++;

            MUTEX_UNLOCK(&stat_lock);
#endif


            /***
             ***
             ***/

            psessp->sid    = t_req;
            psessp->psid   = this->sid;
            psessp->mode   = MODE_SELF;
            psessp->level  = this->level+1;
            psessp->state  = STATE_THINK;
            psessp->last_action = this->last_action;

            gettimeofday(&(psessp->accept_time), NULL);

            psessp->reported = 0;
            psessp->reason  = -1;

			psessp->hf		= 0;

            psessp->pos     =  0;
            psessp->len     = -1;
            psessp->rcode   = -1;
            psessp->count   = 0;
            psessp->lmd     = -1;
            psessp->expire  = -1;

            psessp->cachable = SESS_CACHABLE;
            psessp->reload   = SESS_NORELOAD;
            psessp->ishit    = SESS_MISS;
            psessp->publish  = SESS_PUBLIC;

            psessp->oid     = -1;

            psessp->cfd     = -1;
            psessp->sfd     = -1;
            psessp->rfd     = -1;
            psessp->rcount  = 2;


            strcpy(psessp->clientname, "self");

            memset(psessp->loc , 0, sizeof(psessp->loc));
            memset(psessp->reqbuf, 0, sizeof(psessp->reqbuf));
            psessp->reqlen = 0;

            psessp->cver_major = 0;
            psessp->cver_minor = 0;
            psessp->sver_major = -1;
            psessp->sver_minor = -1;
            psessp->ims = -1;

            DoSetReq(psessp, nurl);

#if 0
            dump_stringchunk("generated request", psessp->reqbuf);
#endif

#ifdef TRACE_PREFETCH
            Trace("nurl = %4d-%02d %p (#%d) '%s'\n",
                this->sid, nc, psessp, psessp->sid, psessp->loc);
#endif


            chk = DoThink(psessp);
            if(chk==0 && psessp->state==STATE_QUERY_ADDR) {
                DoFindServer(psessp);
                if(psessp->state==STATE_CONN_SERV) {
#if 0
                    Trace("connect prefetch target '%s'\n", psessp->loc);
#endif
                    DoServerConnect(psessp);
                }

#ifdef TRACE_PREFETCH
                Trace("sess# %d, <self>   method <%s>,URL '%s' version ?\n",
                    psessp->sid, MSG_M_[psessp->method], psessp->loc);
#else
#ifdef TRACE_SESS
                Trace("sess# %d, <self>   method <%s>,URL '%s' version ?\n",
                    psessp->sid, MSG_M_[psessp->method], psessp->loc);
#endif
#endif

#if 0
                Trace("     push to think Q as sess# %d %s\n",
                    psessp->sid, MSG_STATE_[psessp->state]);
#endif
                if(psessp) {
                    chk = push_sess(think_queue, psessp);
                    if(chk) {
                        Error("DoParseAnchor: sess# %d, fail push to think Q",
                            psessp->sid);
                        sq_be_free(psessp);
                        psessp = NULL;
                    }
                    psessp = NULL;

#if 0
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

                }
            }
            else {
                DoReleaseSess(psessp);
                psessp = NULL;
            }

#if 0
            Trace("  ..done\n");
#endif

skip:
            (void)1;    /* dummy for label */
#if 0
            Trace("  skip\n");
#endif

        }
        else {
            /*
            Trace("DoParseAnchor: fail cache_have(), return %d\n", ihave);
            */
        }
next:
#if 0
        Trace("  next\n");
#endif

        aurl++;
        nc++;
#if 0
        Trace("  ? next\n");
#endif
    }

#ifdef TRACE_PREFETCH
    Trace("DoParseAnchor<%d>: loop end\n");
#endif

    HTML_FreeURLList(urllist);
    FREE(urllist);
    urllist = NULL;

    tags = NULL;


#ifdef TRACE_PREFETCH
    Trace("DoParseAnchor<%d>: block end\n");
#endif

    goto out;

done:

not_found:
#if 0
    /*
    Error("DoParseAnchor: cache was not found #%d [%d:%d] when hit",
        this->sid, this->fhash, this->shash);
    */
    Error("DoParseAnchor: sess#%d [%d:%d:%d] does not have cache entry",
        this->sid, this->fhash, this->shash, this->thash);
#endif
    goto out;


out:

out_nothing:

    return 0;


out_pending:

    return 1;

}




/*
 * Handler
 *
 */

void*
Handler(sess_t *sessp, time_t now)
{
    int chk;
    int cv;
    int pending;

    int m_sid;

    if(sessp==NULL)
        return NULL;


    m_sid = sessp->sid;


#ifdef TRACE_SCHEDULE
    Trace("Handler<%p> start, sess %p, sess# %d, mode %d, state %s\n",
        pthread_self(),
        sessp, sessp->sid, sessp->mode, MSG_STATE_[sessp->state]);
#endif

#if 0
    if(sessp->rcode<0) {
        Trace("sess# %7d %14s %3d\n",
            sessp->sid, MSG_STATE_[sessp->state], sessp->rcode);
    }
#endif

    sessp->last_action = now;
    sessp->tic++;


    switch(sessp->state) {
    /*
     * like manager
     */
    case STATE_ACCEPT:
    case STATE_RECV_REQ:

        if(sessp->mode==MODE_CLIENT && sessp->cfd>=0) {
#if 0
            Trace("client request (%d) and client fd (%d) is normal\n",
                    sessp->mode, sessp->cfd);
#endif

            if(sessp->reqlen<4) {
                chk = DoReadReqFirst(sessp);
            }
            else
            if(sessp->reqlen > L_REQBUF) {
                Error("request is too large");
                sessp->reason = R_TOOLONG_URL;
                goto ignore;
            }
            else {
                chk = DoReadReqRest(sessp);
            }

            if(sessp->state==STATE_MESSAGE)
                goto ignore;

            if(chk) {
#if 1
                Error("handler: fail client request parsing");
#endif
                goto fail;
            }
        }
        else {
#if 0
            Trace("self request (%d) or client fd (%d) is too small\n",
                    sessp->mode, sessp->cfd);
#endif
        }

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 5= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        if(sessp->state!=STATE_THINK) {
            break;
        }
#if 0
        Trace("falling THINK sess# %d,\n", sessp->sid);
#endif

    case STATE_THINK:
#if 0
        Trace("sess# %d, call think()\n", sessp->sid);
#endif
        chk = DoThink(sessp);
        Trace("after DoThink: state %d %c\n",sessp->state,sessp->dhr_state);
        if(chk) {
            if(chk==-1) {
                sessp->reason = R_SOME_FATAL_ERROR;
                sessp->state = STATE_MESSAGE;
                goto ignore;
            }

            Error("Handler: fail session dicision");
            goto fail;
        }

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 6= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        switch(sessp->state) {
        case STATE_REPORT:
            DoReport(sessp);
            DoCloseSess(sessp);
            DoReleaseSess(sessp);
            sessp = NULL;
            goto out;

        case STATE_WAIT_FRIEND:

#ifdef USE_WQUEUE
            /* THINK -> WAIT_FRIEND */
/*
            Trace("TH->WF, sess# %d, enter wait Q %d/%d\n",
                sessp->sid, wait_queue->count, wait_queue->cmax);
*/
#if 0
            chk = push_sess(wait_queue, sessp);
#endif
            chk = sq_timedpush(wait_queue, sessp, 5000);
            if(chk) {
                goto conn;
            }
            sessp = NULL;

            goto out;
#else
/*
            Trace("TH->WF\n");
*/
#endif
            break;

        case STATE_QUERY_ADDR:
            DoFindServer(sessp);

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 21= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
            if(sessp->state==STATE_CONN_SERV) {
                DoServerConnect(sessp);
#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 22= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
            }

            break;

        case STATE_SEND_CACHE:
            DoSendCache(sessp);
/*            Trace("after DoSendCache1: state %d %c\n",sessp->state,sessp->dhr_state);*/

            if(sessp->state==STATE_QUERY_ADDR) {
#if 0
Trace("sess# %d, change to fetch\n", sessp->sid);
#endif
                goto start_fetch;
            }
            break;

        case STATE_MESSAGE:
            goto ignore;

 /*NOTREACHED*/
            break;

        case STATE_DONE:
        case STATE_CLOSE:
            goto done;

#if 1
        /* XXX */
        case STATE_THINK:
/*
            Trace("TH -> TH sess# %d, \n", sessp->sid);
*/
            goto conn;
#endif

	case STATE_DHR_SECONDARY_MISS:
	    DoSendMessage(sessp);
            sessp->state = STATE_CLOSE;
            goto done;

        default:
            Error("Handler: sess# %d, strange state (%d) after DoThink() ",
                sessp->sid, sessp->state);
            goto fail;
        }

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 7= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        break;


    case STATE_CLOSE:
    case STATE_DONE:
        goto done;

 /*NOTREACHED*/
        break;

    case STATE_WAIT_FRIEND:
        DoCheckEndFriendSess(sessp);

#ifdef USE_WQUEUE
        if(sessp->state==STATE_WAIT_FRIEND) {
Trace("WF->WF, re-push sess# %d, to wait Q\n", sessp->sid);


#if 0
            chk = push_sess(wait_queue, sessp);
#endif
            chk = sq_timedpush(wait_queue, sessp, 5000);
            if(chk) {
                goto conn;
            }

            goto out;
        }
        else {
Trace("WF -> %s sess# %d,\n", MSG_STATE_[sessp->state], sessp->sid);
        }
#endif

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 93= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        break;

    case STATE_SEND_CACHE:
        DoSendCache(sessp);
/*            Trace("after DoSendCache2: state %d %c\n",sessp->state,sessp->dhr_state);*/

        if(sessp->state==STATE_QUERY_ADDR) {
#if 0
Trace("sess# %d, change to fetch\n", sessp->sid);
#endif
            goto start_fetch;
        }

        break;

    case STATE_QUERY_ADDR:
start_fetch: (void)1;

        DoFindServer(sessp);
#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 11= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        if(sessp->state==STATE_CONN_SERV) {
            DoServerConnect(sessp);
#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 12= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        }

        break;

    case STATE_WAIT_ADDR:

        chk = hcache_recv_response_addr_direct(sessp->servername,
            &(sessp->serveraddr), sessp->rfd);
        sessp->rfd = -1;
        if(chk) {
            Error("fail hcache_recv_response_addr_direct() to '%s', ret %d <%d>",
                sessp->servername, chk, sessp->rcount);

            sessp->rcount--;
            if(sessp->rcount>0) {
                sessp->state = STATE_QUERY_ADDR;
                goto start_fetch;
            }

            sessp->reason = R_RESOLVER_ERROR;
            sessp->state = STATE_MESSAGE;
        }
        else {
#ifdef TRACE_HCACHE
            Trace("  got address via hcache_recv_response_addr_direct()\n");
#endif
            chk = hcache_set_addr_incache(sessp->servername, &(sessp->serveraddr));

#if 0
            hcache_list();
#endif

            sessp->state = STATE_CONN_SERV;
        }

        if(sessp->state!=STATE_CONN_SERV) {
            goto ignore;
        }

    case STATE_CONN_SERV:

        DoServerConnect(sessp);

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 10= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        break;

#if 0
    case STATE_WAIT_CONN_SERV:

        DoWaitServerConnect(sessp);
        break;
#endif

    case STATE_WAIT_CONN_SERV:
    case STATE_SEND_REQUEST:

Trace("passed DoSendRequest: errno %d state %d\n",errno,sessp->state);
        DoSendRequest(sessp);

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 13= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif
        break;

    case STATE_RECV_RES:

        DoServerReply(sessp);

/*
        Trace("RECV_RES -> %s, sess# %d, \n",
            MSG_STATE_[sessp->state], sessp->sid);
*/

#if 0
if(sessp->cachable==SESS_UNCACHABLE) {
    Trace(" 9= sess# %d, %s\n", sessp->sid, sessp->loc);
    dump_stringsegment("req.", sessp->reqbuf, sessp->reqlen);
}
#endif

        break;

    case STATE_RELAY_DIRECT:

        DoRelayDirect(sessp);
        break;


    default:
        Error("Handler: strange state (%d) ", sessp->state);
        goto fail;

    } /* end of state switch */


    if(sessp!=NULL &&
        sessp->state==STATE_MESSAGE) {
Trace("passed goto ignore\n");
        goto ignore;
    }


done:

    if(sessp!=NULL &&
        (sessp->state==STATE_CLOSE || sessp->state==STATE_DONE)) {

#if 0
    Trace("Handler: done, sess# %d, re-push to think Q\n", sessp->sid);
#endif

        DoCloseSess(sessp);

        chk = MUTEX_LOCK(&config_lock);
        cv = prefetch_flag;
        chk = MUTEX_UNLOCK(&config_lock);

#ifdef TRACE_PREFETCH
        if(!cv) {
            Trace("Handler: prefetch is disabled\n");
        }
        else
#endif

        if(cv &&
            (sessp->level<2 && sessp->rcode>=200 && sessp->rcode<=399) ) {
/*
        sessp->sread > 8
        sessp->swrite > 10
 */

            pending = DoParseAnchor(sessp);
            if(pending) {
                Trace("Handler: pending sess# %d\n", sessp->sid);
            }
        }
#ifdef TRACE_PREFETCH
        else {
            Trace("Handler: sess# %d, skip parse object content [level %d, rcode %d]\n",
                sessp->sid, sessp->level, sessp->rcode);
        }
#endif

        MUTEX_LOCK(&stat_lock);
        process_request_count++;
        MUTEX_UNLOCK(&stat_lock);

        release_buf(sessp);

        DoReleaseSess(sessp);
        sessp = NULL;

        goto out;
    }


    goto conn;

ignore:
#if 0
    Trace("Handler: ignore, sess# %d, close.\n", sessp->sid);
#endif


Trace("passed DoSendMessage\n");
    DoSendMessage(sessp);
    sessp->state = STATE_CLOSE;

    DoCloseSess(sessp);
    DoReleaseSess(sessp);
    sessp = NULL;

    goto out;

fail:
#if 0
    Trace("Handler: fail, sess# %d, close.\n", sessp->sid);
#endif


    DoCloseSess(sessp);
    DoReleaseSess(sessp);
    sessp = NULL;


    goto out;

conn:
    if(sessp!=NULL) {

#if 0
#ifdef TRACE_SCHEDULE
    Trace("Handler<%p> end,   sess %p, sess# %d, mode %d, state %s\n",
        pthread_self(),
        sessp, sessp->sid, sessp->mode, MSG_STATE_[sessp->state]);
#endif
#endif


#if 0
        Trace("### A\n");
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

#if 0
        Trace("Handler: conn, sess# %d, re-push to think Q\n", sessp->sid);
#endif


        chk = push_sess(think_queue, sessp);
        if(chk) {
            Error("Handler: sess# %d, fail re-push to think Q", sessp->sid);

            sessp->reason = R_MEMORY_OVERFLOW;
            DoSendMessage(sessp);
            DoCloseSess(sessp);
            DoReleaseSess(sessp);
            sessp = NULL;
        }
        else {
            sessp = NULL;
        }

#if 0
        Trace("### B\n");
        MUTEX_LOCK(&think_queue->lock);
        fflush(stderr);
        sq_show(stdout, think_queue);
        fflush(stdout);
        MUTEX_UNLOCK(&think_queue->lock);
#endif

    }

    goto out;

out:
    chk = 0;

#if 0
    Trace("### C\n");
    MUTEX_LOCK(&think_queue->lock);
    fflush(stderr);
    sq_show(stdout, think_queue);
    fflush(stdout);
    MUTEX_UNLOCK(&think_queue->lock);
#endif


#ifdef TRACE_SCHEDULE
    if(sessp) {
    Trace("Handler<%p> end,   sess %p, sess# %d, mode %d, state %s\n",
        pthread_self(),
        sessp, sessp->sid, sessp->mode, MSG_STATE_[sessp->state]);
    }
    else {
    Trace("Handler<%p> end,   lost, sess# %d, \n",
        pthread_self(), m_sid);
    }

    Trace("Handler<%p> nhot %d, think queue %d/%d, wait_queue %d/%d\n",
        pthread_self(), nhot, think_queue->count, think_queue->cmax,
        wait_queue->count, wait_queue->cmax);
#endif

    return NULL;
}

inline
int
check_timeout(sess_t *sessp, time_t now)
{
    int ret;
    int idle, elap;

    ret = 0;

    idle = now - sessp->last_action;
    elap = now - sessp->accept_time.tv_sec;
Trace("check_timeout: sess# %d, idle %d, elap %d, state %d\n", sessp->sid, idle, elap, sessp->state);

    switch(sessp->state) {
    case STATE_ACCEPT:
    case STATE_RECV_REQ:
#if 0
    Trace("check_timeout: sess# %d, now %d, idle %d, elap %d, reqlen %d\n",
        sessp->sid, now, idle, elap, sessp->reqlen);
#endif
        if(sessp->last_action == sessp->accept_time.tv_sec) {

            if(sessp->reqlen>0) {
#if 0
    Trace("    Is it empty request ?\n");
#endif
            }

            if(idle > RECEPT_TIMEOUT*3) {
                sessp->reason = R_CLIENT_TIMEOUT;
                ret = 1;
            }
        }
        else
        if(idle > RECEPT_TIMEOUT) {
            sessp->reason = R_CLIENT_TIMEOUT;
            ret = 1;
        }
        break;
    case STATE_WAIT_FRIEND:
        if(idle > WFRIEND_TIMEOUT) {
            sessp->reason = R_WFRIEND_TIMEOUT;
            ret = 1;
        }
        break;
    case STATE_WAIT_CONN_SERV:
        if(idle > FETCH_TIMEOUT) {
            if((sessp->dhr_state == 'A') ||
               (sessp->dhr_state == 'B') ||
               (sessp->dhr_state == 'M'))
                sessp->reason = R_DHR_SIBLING_FAILURE;
            else
                sessp->reason = R_SERVER_TIMEOUT;
            ret = 1;
        }
        break;
    case STATE_RECV_RES:
        if(idle > RECV_RES_TIMEOUT) {
                sessp->reason = R_SERVER_TIMEOUT;
            ret = 1;
        }
    default:
        if(idle > ELSE_TIMEOUT) {
            sessp->reason = R_TIMEOUT;
            ret = 1;
        }
        break;
    }

    if(ret==0) {
        if(elap > SESSION_TOTAL_TIMEOUT) {
            sessp->reason = R_TIMEOUT;
            ret = 1;
        }
    }

#if 0
if(ret!=0) {
    Trace("sess# %d, elap/idle %d/%d -> reason %d, ret %d\n",
        sessp->sid, elap, idle, sessp->reason, ret);
}
#endif

    return ret;
}



#if 0
#define CHECK {\
    int k;\
    Trace("w_sess_count %d, w_fd_count %d\n", w_sess_count, w_fd_count);\
    if(w_sess_count>0) {\
        Trace("index sess     sid#  state           |cfd   sfd\n");\
        for(k=0;k<w_sess_count;k++) {\
            if(sess_chunk[k]) \
                Trace("%5d %8x %5d %-15s |%5d %5d \n",\
                    k, sess_chunk[k],\
                    sess_chunk[k]->sid, MSG_STATE_[sess_chunk[k]->state],\
                    sess_chunk[k]->cfd, sess_chunk[k]->sfd\
                    );\
            else\
                Trace("%5d empty\n", k);\
        }\
    }\
    else {\
        Trace("***no sessions\n");\
    }\
    if(w_fd_count>0) {\
        Trace("\nindex fd    |sess   sid# state\n");\
        for(k=0;k<w_fd_count;k++) {\
            if(bound[k]) \
                Trace("%5d %5d |%8x %5d %-15s\n",\
                    k, fdrec[k].fd,\
                    bound[k], bound[k]->sid, MSG_STATE_[bound[k]->state]);\
            else\
                Trace("%5d %5d empty\n", k, fdrec[k].fd);\
        }\
    }\
    else {\
        Trace("***no fds\n");\
    }\
    Trace("= = \n");\
}

#else
#define CHECK   ((void)0)
#endif

static
int
sess_comp(const void *ap, const void *bp)
{
    sess_t *a = *(sess_t**)ap;
    sess_t *b = *(sess_t**)bp;

    if(a->fhash==b->fhash) {
        return a->shash - b->shash;
    }
    else {
        return a->fhash - b->fhash;
    }
}



void*
Slave(void *dummy)
{
    time_t now;
    int w_sess_max;
    int w_sess_count;
    int w_fd_count;
    int timeout_count;
    sess_t *sess_chunk[SESS_CHUNK_SIZE];
    struct pollfd fdrec[SESS_CHUNK_SIZE*2];
    sess_t *bound[SESS_CHUNK_SIZE*2];
    int bk_use[RADIX_H1_MAX];
    int i, j;
    int chk;
    int try;
    int touched;
    int same;

    int focus;
    int c;

    int miss_capability=4;

    int sq_show_w_lock(FILE*, sess_queue_t*);

    int w;


#if 0
Trace("Slave: pr %d, th %d\n", getpid(), pthread_self());
#endif

    w_sess_max = require_chunk_size;

    /* forever loop */
    while(1) {
        timeout_count = 0;
        w_sess_count = 0;
        w_fd_count = 0;

        for(i=0;i<SESS_CHUNK_SIZE;i++) {
            sess_chunk[i] = NULL;
            bound[i] = NULL;
        }
        for(i=0;i<RADIX_H1_MAX;i++) {
            bk_use[i] = 0;
        }

        w_sess_count = sq_pullchunk(think_queue, sess_chunk, w_sess_max);

        if(w_sess_count<=0) {
            goto next_turn;
        }
        now = time(NULL);

        /*
         *
         */
        for(i=0;i<w_sess_count;i++) {
            if(sess_chunk[i]) {
                if(sess_chunk[i]->fhash>=0) {
                    bk_use[sess_chunk[i]->fhash]++;
                }
            }

            if(sess_chunk[i]->last_action>0
                && check_timeout(sess_chunk[i], now)) {
Trace("passed check_timeout in Slave  dhr_state:%c\n",sess_chunk[i]->dhr_state);

#if 1
                /* XXX */
                if(sess_chunk[i]->reason==R_WFRIEND_TIMEOUT) {
Trace("sess# %d, switch wait FRIEND to retrieve\n", sess_chunk[i]->sid);
Log("sess# %d, switch wait FRIEND to retrieve", sess_chunk[i]->sid);
/*
*/

                    sess_chunk[i]->state = STATE_QUERY_ADDR;
                    sess_chunk[i]->cachable = SESS_UNCACHABLE;

                    continue;
                }
#endif

                if(sess_chunk[i]->reason == R_DHR_SIBLING_FAILURE){
                    DoHalfServerCloseSess(sess_chunk[i]);
                    if(sess_chunk[i]->dhr_state == 'A'){
                        sess_chunk[i]->dhr_hash = 
                          dhr_hash2(sess_chunk[i]->servername, 
                                    sess_chunk[i]->serverpath, n_proxy); 
                        sess_chunk[i]->dhr_state = 'S';
                    }
                    if(sess_chunk[i]->dhr_state == 'B'){
                        sess_chunk[i]->dhr_hash = 
                          dhr_hash1(sess_chunk[i]->servername, 
                                    sess_chunk[i]->serverpath, n_proxy); 
                        sess_chunk[i]->dhr_state = 'S';
                    }
                    if(sess_chunk[i]->dhr_state == 'M'){
                        sess_chunk[i]->dhr_hash = -1;
                        sess_chunk[i]->dhr_state = 'E';
                    }
                    sess_chunk[i]->state = STATE_QUERY_ADDR;
                    continue;
                }

                release_buf(sess_chunk[i]);
                DoTimeoutClose(sess_chunk[i]);
                DoReleaseSess(sess_chunk[i]);
                timeout_count++;

                sess_chunk[i] = NULL;


                continue;
            }


            switch(sess_chunk[i]->state) {
            case STATE_IDLE:
                break;
            case STATE_ACCEPT:
            case STATE_RECV_REQ:
                fdrec[w_fd_count].fd = sess_chunk[i]->cfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                fdrec[w_fd_count].events |= POLLERR;
/*
*/
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_WAIT_ADDR:
                fdrec[w_fd_count].fd = sess_chunk[i]->rfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                fdrec[w_fd_count].events |= POLLERR;
/*
*/
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_WAIT_CONN_SERV:
                fdrec[w_fd_count].fd = sess_chunk[i]->sfd;
                fdrec[w_fd_count].events = POLLOUT | POLLWRNORM;
                fdrec[w_fd_count].events |= POLLERR;
/*
*/
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_RECV_RES:
                fdrec[w_fd_count].fd = sess_chunk[i]->sfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                fdrec[w_fd_count].events |= POLLERR;
/*
*/
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            case STATE_RELAY_DIRECT:
                fdrec[w_fd_count].fd = sess_chunk[i]->cfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                fdrec[w_fd_count].events |= POLLERR;
/*
*/
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                fdrec[w_fd_count].fd = sess_chunk[i]->sfd;
                fdrec[w_fd_count].events = POLLIN | POLLRDNORM;
                fdrec[w_fd_count].events |= POLLERR;
/*
*/
                bound[w_fd_count] = sess_chunk[i];
                w_fd_count++;
                break;
            default:
#ifdef TRACE_SCHEDULE
                Trace("Slave: thread %x, sess# %d, mode %c, state strange %s\n",
                    pthread_self(),
                    sess_chunk[i]->sid,
                    sess_chunk[i]->mode==MODE_CLIENT ? 'c': 's',
                    MSG_STATE_[sess_chunk[i]->state]);
#endif
                break;
            }
        }


#ifdef TRACE_SCHEDULE
        Trace("th %d, w_sess_count/max %d/%d w_fd_count %d\n",
            pthread_self(), w_sess_count, w_sess_max, w_fd_count);

        /*
        Trace("\nbefore ***\n");
        CHECK;
        */
#endif


        if(w_fd_count<1) {
Trace("passed goto dispatch_by_sess: 5986 errno %d\n",errno);
            goto dispatch_by_sess;
        }

        c = 0;
try_poll:
        c++;
        chk = poll(fdrec, w_fd_count, 1000); /* 1000msec = 1sec */

#ifdef TRACE_SCHEDULE
        if(timeout_count) {
            Trace("Slave: <%08x> query/receive/timeout %4d/%4d/%4d\n",
                pthread_self(), w_fd_count, chk, timeout_count);
        }
#endif

        if(chk<=0) {
          if(chk<0) {
            if(errno==EINTR && c<24) {
                goto try_poll;
            }
            Error("fail poll() %d (%d)", c, errno);
            if(errno==ENOMEM) {
                msleep(1);
            }
          }

            for(i=0;i<w_fd_count;i++) {
                if(bound[i]) {
                    chk = push_sess(think_queue, bound[i]);
                    if(chk) {
                        bound[i]->reason = R_MEMORY_OVERFLOW;

                        DoSendMessage(bound[i]);
                        DoCloseSess(bound[i]);
                        release_buf(bound[i]);
                        DoReleaseSess(bound[i]);
                    }
                    for(j=0;j<w_sess_count;j++) {
                        if(sess_chunk[j]==bound[i]) {
                            sess_chunk[j] = NULL;
                        }
                    }
                    bound[i] = NULL;
                }
            }

Trace("passed goto dispatch_by_sess: 6033 errno %d\n",errno);
            goto dispatch_by_sess;
        }


dispatch_by_fd:
        for(i=0;i<w_fd_count;i++) {
            touched = 0;

            /* count other fds is bounded to same session */
            same = 0;
            for(j=i+1;j<w_fd_count;j++) {
                if(bound[j]==bound[i]) {
                    same++;
                }
            }

            if(bound[i]==NULL) {
                continue;
            }

            /* server connect when state WAIT_CONN_SERV
             * otherwise, receive anything from server or shutdown arrival
             */
            if(
                (bound[i]->state==STATE_WAIT_CONN_SERV
                    && fdrec[i].revents & (POLLOUT | POLLERR) )
                  ||(bound[i]->state!=STATE_WAIT_CONN_SERV
                    && fdrec[i].revents & (POLLHUP | POLLIN) )
                ) {
Trace("Slave b: errno %d\n",errno);
#ifdef TRACE_SCHEDULE
                Trace("  call handler<%p> %d: fd# %d, sess %p, sess# %d, %s\n",
                    pthread_self(), i, fdrec[i].fd,
                    bound[i], bound[i]->sid,
                    MSG_STATE_[bound[i]->state]);
#endif
#if 0
                if(fdrec[i].revents & POLLHUP) {
                    Trace("got POLLHUP sess# %d, %s\n",
                        bound[i]->sid, MSG_STATE_[bound[i]->state]);
                }
#endif
Trace("passed Handler in Slave 1: errno %d\n",errno);
                Handler(bound[i], now);
                touched = 1;
            }
#if 1
            else
            if(fdrec[i].revents & POLLERR) {
                    Trace("got POLLERR sess# %d, %s\n",
                        bound[i]->sid, MSG_STATE_[bound[i]->state]);
            }
#endif
            else {
                if(same==0) {
                    /* nobody is bounded to same session */
                    chk = push_sess(think_queue, bound[i]);
                    if(chk) {
                            bound[i]->reason = R_MEMORY_OVERFLOW;
                            DoSendMessage(bound[i]);
                            DoCloseSess(bound[i]);
                            DoReleaseSess(bound[i]);
                            bound[i] = NULL;
                    }
                }
            }

            /* clear a processed session */
            for(j=0;j<w_sess_count;j++) {
                if(bound[i]==sess_chunk[j]) {
                    sess_chunk[j] = NULL;
                }
            }
            if(touched && same) {
                for(j=i+1;j<w_fd_count;j++) {
                    if(bound[j]==bound[i]) {
                        bound[j] = NULL;
                    }
                }
            }
            bound[i] = NULL;
        }


dispatch_by_sess:


        for(i=0;i<w_sess_count;i++) {
            if(sess_chunk[i]) {
#ifdef TRACE_SCHEDULE
                Trace("  call handler<%p> %d: sess %p, sess# %d, %s\n",
                    pthread_self(), i,
                    sess_chunk[i], sess_chunk[i]->sid,
                    MSG_STATE_[sess_chunk[i]->state]);
#endif
                Handler(sess_chunk[i], now);

                for(j=0;j<w_sess_count;j++) {
                    if(i!=j && sess_chunk[i]==sess_chunk[j]) {
                        sess_chunk[j] = NULL;
                    }
                }
                sess_chunk[i]=NULL;
            }
        }


next_turn:
        (void)0; /* dummy */

        /*
         *
         */
        
#if 0
 {
        char line[K(20)];
        int c;

        c = 0;
        for(i=0;i<radix_h1;i++) {
            if(bk_use[i]>0) {
                c++;
                line[i] = '*';
            }
            else {
                line[i] = '_';
            }
        }
        line[i] = '\0';
        Trace("bkuse %3d %s\n", c, line);
 }
#endif





        /*** 
         *** clean-up bucket on memory and directories in disk
         ***/

        now = time(NULL);

        for(i=0;i<radix_h1;i++) {
            bucket_t *bk;
            int chk;

            if(bk_use[i]<=0) {
                continue;
            }

            bk = &skcache_bucket_base[i];
            TRYLOCK_BK(bk, chk);
/*
            Trace("check timming for clean-up bk# %d, chk=%d\n",
                bk->bid, chk);
*/
            if(chk==EBUSY) {
                continue;
            }
#if 0
 Trace("  try-lock bk# %d, chk=%d\n", bk->bid, chk);
#endif

#if 0
Trace("check reaper calling bk# %d, entry u %d h %d ; nt %d ; p %d ; ck %d range %d ; diff %d O\n",
    bk->bid,
    bk->entry_use, bk->entry_hwater, bk->gc_ntarget,  bk->gc_push,
    bk->gc_checkpoint, now - skcache_mem_keep_cur,
    bk->gc_checkpoint - (now - skcache_mem_keep_cur));
#endif
            if(bk->entry_use > bk->entry_hwater
                || bk->gc_ntarget>0 || bk->gc_push>0
                || (bk->gc_checkpoint>0 &&
                    bk->gc_checkpoint - ( now - skcache_mem_keep_cur ) < 0 ) )
            {

#ifdef TRACE_REAPER
Trace("check reaper calling bk# %d, entry u %d h %d ; nt %d ; p %d ; ck %d range %d ; diff %d I\n",
    bk->bid,
    bk->entry_use, bk->entry_hwater, bk->gc_ntarget,  bk->gc_push,
    bk->gc_checkpoint, now - skcache_mem_keep_cur,
    bk->gc_checkpoint - (now - skcache_mem_keep_cur));
#endif

#if 0
                if(bk->gc_checkpoint>0 &&
                    bk->gc_checkpoint - ( now - skcache_mem_keep_cur ) > 0 ) {
                    bk->gc_ntarget = 0;
                }
#endif

                bk->own = 'M';
                chk = reaper(bk, bk->entry_all/20);
                bk->own = ' ';

                bk->gc_push = 0;
            }

#ifdef USE_3LEVEL_DIR
            bk->dirpos = (bk->dirpos + 1 ) %(radix_h2*radix_h3);
#else
            bk->dirpos = (bk->dirpos + 1 ) %radix_h2;
#endif
            focus = bk->dirpos;
#if 0
            Trace("  try check dir. bk# %d, focus %d\n", bk->bid, focus);
#endif

            UNLOCK_BK(bk);
        


            /***
             *** sweep directories
             ***/

          {
            int p, q, r, qrange;
            int e2, e1, e3;
            int dcount, dtime;
            long long dsize;
            time_t now;
            int difftime;

            now = time(NULL);

#ifdef USE_3LEVEL_DIR
            qrange = radix_h2*radix_h3;
#else
            qrange = radix_h2;
#endif

            q = focus;
            r = -1;

            e1 = bk->bid;
#ifdef USE_3LEVEL_DIR
            e2 = q / radix_h3;
            e3 = q % radix_h3;
#else
            e2 = q;
            e3 = 0;
#endif
            p = DSIG_POS(e1, e2, e3);

            MUTEX_LOCK(&dir_mutex);
            dtime   = dir_time[p];
            dcount  = dir_count[p];
            dsize   = (long long) dir_sz[p];

            /* not GCed */
            if(dtime<0) {
                difftime = now - boot_date;
            }
            else {
                difftime = now - dtime;
                if(difftime >= DSIG_MIN_GAP && dsize > dir_size_hwater
                    && dir_lock[p]<=0) {
                    r = p;
                    goto dir_found;
                }
            }
#if 0
            /* add random factor */
            difftime += rand()%(skcache_signature_refresh/5);
#endif

            if(difftime >= skcache_signature_refresh
                && dir_lock[p]<=0) {
                r = p;
                goto dir_found;
            }
dir_found:
        
            if(r<0) {
                MUTEX_UNLOCK(&dir_mutex);
            }
            else {
#ifdef TRACE_CHKDIR
Trace("  call chkdir dir# %d, [%d:%d:%d] time %d [%d,%d] %df %lldB %.2fKB / %.2fKB\n", 
                    r, e1, e2, e3,
                    difftime, DSIG_MIN_GAP, skcache_signature_refresh,
                    dcount,
                    (long long)dsize, fdK(dsize), fdK(dir_size_hwater));
#endif
                dir_lock[r] = pthread_self();
                MUTEX_UNLOCK(&dir_mutex);

                chk = skcache_check_dir(e1, e2, e3, 1);

                MUTEX_LOCK(&dir_mutex);
                dir_lock[r] = 0;
                MUTEX_UNLOCK(&dir_mutex);
            }
          }



        } /* for, clean-up loop */

    } /* while, forever loop */

}



void*
Reporter(void *dummy)
{
    int cfd;
    int w;
    int p;

    while(1) {
        AcceptReport_fp(stdout);
        ListSession(stdout);
#if 0
        sim_stat_show_lock(&len_perread_stat, stderr);  /* XXX */
#endif
#if 1
        sim_stat_show_lock(&object_size_stat, stderr);  /* XXX */
        cache_list_entry(stderr); /* XXX */
#endif
        sleep(900);
    }
}


extern volatile int disk_sweep_interval;

void*
DSweeper(void* dummy)
{
    int chk;
    struct timespec join_time;

    /* msec, how many times in 1000msec */
    int interval = disk_sweep_interval;

#if 0
Trace("DSweeper: pr %d, th %d\n", getpid(), pthread_self());
#endif

    while(1) {
#ifdef USE_3LEVEL_DIR
        skcache_dir_sweeper_level3();
#else
        skcache_dir_sweeper();
#endif

#if 0
        join_time.tv_sec    = 0;
        join_time.tv_nsec   = interval * 1000000;
        chk = nanosleep(&join_time, NULL);
#endif
        chk = msleep(interval);

        /*
        Trace("DSweeper: chk %d\n", chk);
        */
    }
}



static
int
check_mutex(pthread_mutex_t *mx)
{
    int chk;

    chk = MUTEX_TRYLOCK(mx);
    if(chk) {
        if(chk==EBUSY)
            return 0;
        else
            return -chk;
    }
    MUTEX_UNLOCK(mx);
    return 1;
}


static char *_msg_idle  = "IDLE ";
static char *_msg_busy  = "BUSY ";
static char *_msg_err   = "ERROR";

static
int
report_check_mutex(FILE *fp, char *msg, pthread_mutex_t *mx)
{
    int chk;
    char *rep;

    chk = check_mutex(mx);
    if(chk<0)   rep=_msg_err;
    if(chk==0)  rep=_msg_busy;
    if(chk>0)   rep=_msg_idle;

    fprintf(fp, "$$$ %s %s\n", msg, rep);

    return 0;
}

void print_sess(FILE *ofp, void* raw_addr, struct timeval *reftime)
{
    sess_t *self;
    int chk, err;
    char *rep;
    struct timeval reftime_local;
    int elap;

    if(reftime==NULL) {
        gettimeofday(&reftime_local, NULL);
        reftime = &reftime_local;
    }

    self = (sess_t*) raw_addr;

    chk = check_mutex(&(self->busy));
    if(chk<0)   rep=_msg_err;
    if(chk==0)  rep=_msg_busy;
    if(chk>0)   rep=_msg_idle;

    elap = reftime->tv_sec - self->accept_time.tv_sec;

    fprintf(ofp, "%3d %s %c %3d %3d <%s>\n",
        self->sid, rep, self->mode==MODE_CLIENT ? 'c': '-', elap, self->fhash, self->loc);
#if 0
    fprintf(ofp, "%3d ", self->reqlen);
    fprintf(ofp, "<%s> ", self->loc);
    fprintf(ofp, "\n");
#endif
}

int
check_buckets()
{
    int i;
    bucket_t *bk;
    int f;

    for(i=0;i<radix_h1;i++) {
        bk = &skcache_bucket_base[i];
        f = check_mutex(&bk->mutex);
        Trace("bk# %4d: %d\n", i, f);
    }

    return 1;
}

int
check_busy_buckets(int *to, int *ms, int *ot)
{
    int i;
    bucket_t *bk;
    int f;
    int c;
    char band[RADIX_H1_MAX];
    char fill[RADIX_H1_MAX];

    *to = 0;
    *ms = 0;
    *ot = 0;
    c = 0;
    for(i=0;i<radix_h1;i++) {
        bk = &skcache_bucket_base[i];

/*
        fill[i] = "_ueX"[(bk->entry_use>0)*1+(bk->gc_checkpoint>0)*2];
*/
        fill[i] = "0123456789X"[10*bk->entry_use/bk->entry_all];

        f = check_mutex(&bk->mutex);
        if(f==0) {
            (*to)++;
#if 0
            if(c==0) {
                Trace("busy buckets:\n");
            }
#endif
#if 0
            Trace("busy bk# %-3d: %d %d\n", i, f, bk->gc_last);
#endif
            c++;
            if(bk->own=='M' || bk->own=='D') {
                (*ms)++;
                band[i] = bk->own;
            }
            else {
                (*ot)++;
                band[i] = '*';
            }
        }
        else
        if(f<0) {
            band[i] = '?';
        }
        else {
            band[i] = '.';
        }
    }
    band[radix_h1] = '\0';
    fill[radix_h1] = '\0';

    if(c>0) {
#if 0
        Trace("busy %d threads\n", c);
#endif
#if 0
    Trace("busy band %d %3d %s\n", time(NULL), c, band);
#endif
    }

#if 1
    Trace("busy band %d %3d %s\n", time(NULL), c, band);
#endif
#if 1
    Trace("busy fill %d %3d %s\n", time(NULL), c, fill);
#endif

    return 1;
}





extern sim_stat_rec hres_stat;



extern long long swap_count;
extern long long swap_size;


extern sim_stat_rec swapload_time;
extern sim_stat_rec swapin_time;
extern sim_stat_rec swapout_time;
extern sim_stat_rec checkdir_time;
extern sim_stat_rec checkdir_time_short;
extern sim_stat_rec compaction_time;

extern sim_stat_rec swapout_fail_reason;
extern sim_stat_rec swapload_fail_reason;



void*
Monitor(void *dummy)
{

/* create */
#define C(x)    int c_##x=0, last_##x=0, delta_##x; \
                double r_##x=-1.0, rmax_##x

/* process */
#define P(x) \
                delta_##x = c_##x - last_##x; \
                r_##x = ((double)delta_##x*1000)/delta_t; \
                if(r_##x>rmax_##x) { \
                    rmax_##x = r_##x; \
                }

/* values */
#define V(x)    r_##x, rmax_##x, (long)c_##x, delta_##x

/* last */
#define L(x)    last_##x = c_##x

    char *fmt =
": time %d %d %s %c\
 arr %.1f/%.1f %ld %d, pro %.1f/%.1f %ld %d, prog %.2f\
 err %.1f/%.1f %ld %d :\
 rtime %.3f mtime %.3f dtime %.3f :\
 hit %d, miss %d, hit%% short %.2f all %.2f,\
 sess %d %d, queue %d %d %d, bk %d %d, buf fail %d success %d :\
 msweep %.1f/%.1f %ld %d, comp %.1f/%.1f %ld %d,\
 swapload %.1f/%.1f %ld %d success %.1f/%.1f %ld %d,\
 swapout %.1f/%.1f %ld %d success %.1f/%.1f %ld %d token %d/%d,\
 frag %d %d %.2f%%,\
 keep %d %d :\
 dsweep bg %.1f/%.1f %ld %d %d %d,\
 chkdir dir %.1f/%.1f %ld %d rfile %.1f/%.1f %ld %d hfile %.1f/%.1f %ld %d,\
 capacity %lld %d %lld"
 ;

    char outline[L_BUF];
    char elap_str[STRING_SIZE];
    struct timeval last_print_t;

    int count;
    int last_keep_depth;
    int keep_check_interval;

    /* time */
    struct timeval last_t, t0, t1, t2, t3, t4, t5, t6;
    int delta_t;
    int t_nhot, t_nfriend;
    char mark;

    int print_count;
    time_t elap;

    /* average in shortime: retrieval, reaper and check dir. */
    double av_rtime;
    double av_reaper;
    double av_chkdir;

    int c_rjq;

    unsigned int sig, last_sig;

    double hrate, hratio;

    int  c_think, c_send, c_fetch, c_prompt;
    int  s_think, s_send, s_fetch, s_prompt;
/*
    int  c_msweep;
*/
    int  c_wait;
    int  c_swap;

    int c_expire_normal_idle_th, c_expire_cracked_idle_th;
#if 0
    int h_expire[COMP_HIST_WIDTH];
#endif
    double expire_k;
    int expire_raise_tick;
    int expire_min;

    double fratio;

    C(arr);             /* arrival      */
    C(pro);             /* process      */
    C(err);             /* error        */
    C(hit);             /* hit          */
    C(miss);            /* miss         */
    C(comp);            /* compaction   */
    C(expire);          /* expiration   */
    C(swapload);        /* swapload         */
    C(swaploadsuc);     /* swapload success */
    C(swapin);          /* swapin           */
    C(swapinsuc);       /* swapin success   */
    C(swapout);         /* swapout          */
    C(swapoutsuc);      /* swapout success  */
    C(dsweep);          /* disk sweep   */
    C(msweep);          /* memoey sweep */
    C(chdcount);        /* check_dir    dir */
    C(chfrcount);       /* check_dir    file */
    C(chfhcount);       /* check_dir    file */

    int kn,ukn;
    long long guess_size;
    long long dir_size_sum;

    int to, ms, ot;




    count = 0;
    print_count = 0;
    last_keep_depth = 0;
    keep_check_interval = 4;

    last_sig = 0xf0f0;

    /* EX TH */
    expire_k = 2.0;
    expire_raise_tick = 60;
    expire_min = 60;


    gettimeofday(&last_t, NULL);
    while(1) {
        /* interval */
        sleep(MONITOR_INTERVAL);

        gettimeofday(&t0, NULL);
        MUTEX_LOCK(&stat_lock);
        elap    = t0.tv_sec - boot_date;
        c_arr   = receive_request_count;
        c_pro   = process_request_count;
        c_err   = error_reply_count;
        c_rjq   = reject_request_count;
        c_comp  = compaction_count;
        c_expire  = expire_count;

        c_hit  = hit_reply_count;
        c_miss = miss_reply_count;

        c_swapload  = swapload_count;
        c_swapin    = swapin_count;
        c_swapout   = swapout_count;
        c_swaploadsuc   = swapload_success;
        c_swapinsuc     = swapin_success;
        c_swapoutsuc    = swapout_success;

        c_dsweep     = dsweeper_count;
        c_msweep     = msweeper_count;

        c_chdcount = checkdir_dir_count;
        c_chfrcount = checkdir_file_read_count;
        c_chfhcount = checkdir_file_handle_count;

        MUTEX_UNLOCK(&stat_lock);

        av_rtime = sim_stat_get_ave(&ret_time_short);
        av_reaper = sim_stat_get_ave(&msweep_time_short);
        av_chkdir = sim_stat_get_ave(&checkdir_time_short);

        MUTEX_LOCK(&sess_lock);
        t_nhot = nhot;
        t_nfriend = nfriend;
        MUTEX_UNLOCK(&sess_lock);

        MUTEX_LOCK(&think_queue->lock);
        c_think = think_queue->count;
        s_think = think_queue->sig;
        MUTEX_UNLOCK(&think_queue->lock);

#ifdef USE_WQUEUE
        MUTEX_LOCK(&wait_queue->lock);
        c_wait = wait_queue->count;
        MUTEX_UNLOCK(&wait_queue->lock);
#else
        c_wait = 0;
#endif

        c_swap = 0;
        c_send   = s_send   = 0;
        c_fetch  = s_fetch  = 0;
        c_prompt = s_prompt = 0;


        gettimeofday(&t6, NULL);

        sig = c_arr | c_pro | t_nhot | s_think | s_send | s_fetch | s_prompt;

        delta_t = KOTETU_SUB_TIME(t0, last_t);

        P(arr);
        P(pro);
        P(err);
        P(hit);
        P(miss);

        hratio = (hit_reply_count+miss_reply_count>0 ?
                    (100.0*hit_reply_count) /(hit_reply_count+miss_reply_count)
                    : -1);
        hrate  = (delta_hit+delta_miss>0 ?
                    (100.0*delta_hit)/(delta_hit+delta_miss)
                    : -1);

        P(comp);
        P(msweep);
        P(dsweep);
        P(swapload);    P(swaploadsuc);
        P(swapin);      P(swapinsuc);
        P(swapout);     P(swapoutsuc);
        P(chdcount);
        P(chfrcount);
        P(chfhcount);

        fratio = ((double)frag_use*100.0)/(frag_use+frag_free);


        MUTEX_LOCK(&config_lock);
        c_expire_normal_idle_th   = skcache_mem_keep_cur;
        c_expire_cracked_idle_th  = -1;
        MUTEX_UNLOCK(&config_lock);


#if 0
        if(t0.tv_sec-last_print_t.tv_sec>=MONITOR_INTERVAL|| sig!=last_sig)
#endif
        if(1)
        {
            if(sig!=last_sig) {
                mark = ':';
            }
            else {
#if 0
s_show_freelist();
#endif
                mark = '*';
            }

            conv_sectoshortstr(elap_str, elap);

            check_busy_buckets(&to, &ms, &ot);

            MUTEX_LOCK(&stat_lock);

#if 1
Trace("Traffic:\n\
    intra_incoming %12lld B %10.2f MB %7.2f GB\n\
    extra_outgoing %12lld B %10.2f MB %7.2f GB\n\
    extra_incoming %12lld B %10.2f MB %7.2f GB\n\
    intra_outgoing %12lld B %10.2f MB %7.2f GB\n",
  (long long)intra_incoming, fdM(intra_incoming), fdG(intra_incoming),
  (long long)extra_outgoing, fdM(extra_outgoing), fdG(extra_outgoing),
  (long long)extra_incoming, fdM(extra_incoming), fdG(extra_incoming),
  (long long)intra_outgoing, fdM(intra_outgoing), fdG(intra_outgoing)
        );
#endif
#if 0
Trace("swap: count %lld (%.2f M), size %lld B %8.2f MB, %5.2f GB\n",
        swap_count, fdM(swap_count),
        swap_size, fdM(swap_size), fdG(swap_size) );
#endif
#if 1
 {
    int i;

    kn = ukn = 0;
    dir_size_sum = 0LL;
    for(i=0;i<dir_num;i++) {
        /* up */
        if(dir_sz[i]>=(off_t)0L) {
            kn++;
            dir_size_sum += dir_sz[i];
        }
        else {
            ukn++;
        }
    }

    if(kn>0) {
        guess_size = ((long long)dir_size_sum*dir_num)/kn;
    }
    else {
        guess_size = -1LL;
    }

Trace("DiskUtil:\n\
    scaned %d dir.(known %u, unknown %u) used size %lld B, %8.2f MB, %5.2f GB\n\
    guessed total used size %lld B, %8.2f MB, %8.2f GB\n",
        dir_num, kn, ukn, dir_size_sum, fdM(dir_size_sum), fdG(dir_size_sum),
        guess_size, fdM(guess_size), fdG(guess_size));

#if 0
    if(ukn<10) {
        Trace("  unknown directories:\n");
        for(i=0;i<dir_num;i++) {
            if(dir_sz[i]<(off_t)0L) {
                Trace("    dir# %d, time %d, size %lld\n",
                    i, dir_time[i], (long long)dir_sz[i]);
            }
        }
    }
#endif


 }
#endif

#ifndef NDEBUG
 {
    int i;
    bucket_t *bk;
    int guess_count;

    guess_count = 0;
    for(i=0;i<radix_h1;i++) {
        bk = &skcache_bucket_base[i];
        
        guess_count += bk->frag_use;
    }

    Trace("Check:\n\
  frag verify: %d : recorded in buckets %d (ave. %.2f) : actual used %d free %d : diff %d\n",
        t0.tv_sec,
        guess_count, (double)guess_count/radix_h1,
        frag_use, frag_free, frag_use - guess_count);
 }
#endif

#if 1
 {
    extern volatile int check_dir_ignore_hash_count;
    extern int dir_conflict;

Trace("  directory conflict: %d\n", dir_conflict);
Trace("  ignore hash count:  %d\n", check_dir_ignore_hash_count);

    sim_stat_show_lock(&swapload_fail_reason,   stderr);
    sim_stat_show_lock(&swapout_fail_reason,    stderr);
    sim_stat_show_lock(&rcode_stat, stderr);
    sim_stat_showX_lock(&req_field_type_stat, stderr);

 }

#endif


            sprintf(outline, fmt,
                t0.tv_sec,
                elap, elap_str,
                mark,

                V(arr),
                V(pro),
                c_arr>0 ?  ((double)c_pro*100.0)/c_arr : -100.0,
                V(err),

                av_rtime,
                av_reaper,
                av_chkdir,
    
                hit_reply_count, miss_reply_count,
                hrate, hratio,

                t_nhot,
                t_nfriend,

                c_think,
                c_wait,
                c_swap,

                to,
                ms,

                alloc_tmp_buf_count,
                alloc_sess_buf_count,

                V(msweep),
                V(comp),
                V(swapload), V(swaploadsuc),
                V(swapout), V(swapoutsuc),

                swapout_token, swapout_token_init,
                frag_use, frag_free, fratio,
                skcache_mem_keep_cur, skcache_entry_keep_cur,

                V(dsweep), dsweeper_phase, dsweeper_pos, 
                V(chdcount),
                V(chfrcount),
                V(chfhcount),
                
#if 0
                swap_count, swap_size
#endif
                dir_size_sum, ukn, guess_size

                );

            MUTEX_UNLOCK(&stat_lock);

            LogLine(outline);
            LogSync();

#ifndef LOG_REPORT_CONSOLE
            Trace("%s\n", outline);
#endif

            sim_stat_reset(&ret_time_short);
            sim_stat_reset(&msweep_time_short);
            sim_stat_reset(&checkdir_time_short);

            sim_hist_add_value(&fratio_hist, (long)fratio);

            skcache_update_keeptime(delta_t);


            last_sig = sig;
#if 0
            last_print_t = t0;
#endif
            memcpy(&last_print_t, &t0, sizeof(struct timeval));
            print_count++;
        }

        {
            sim_hist_add_value(&comp_hist, (long)delta_comp);
            sim_hist_add_value(&expire_hist, (long)delta_expire);

#if 0
            sim_hist_show(&comp_hist, stdout);
            sim_hist_show(&expire_hist, stdout);
#endif

#if 0
            r_comp = sim_hist_average_d(&comp_hist);
            r_expire = sim_hist_average_d(&expire_hist);
            Trace("  $$ rate comp %.2f, expire %.2f\n", r_comp, r_expire);
#endif
        }

#if 1
 /*
  * Monitor 
  */
 if(print_count%30==0) 
 {
    sim_stat_show95_lock(&ret_time,       stderr);
    sim_stat_show95_lock(&msweep_time,        stderr);
    sim_stat_show95_lock(&checkdir_time,    stderr);

    sim_stat_show95_lock(&swapload_time,    stderr);
    sim_stat_show95_lock(&swapout_time,     stderr);
/*
    sim_stat_show95_lock(&swapin_time,      stderr);
*/
 }
#endif

        /* store last values */
        last_t = t0;

        L(arr);
        L(pro);
        L(err);
        L(hit);
        L(miss);
        L(comp);
        L(msweep);
        L(expire);
        L(swapload);
        L(swaploadsuc);
        L(swapout);
        L(swapoutsuc);
        L(dsweep);
        L(chdcount);
        L(chfrcount);
        L(chfhcount);

#if 0
        check_checkpoints();
#endif

        count++;
    }


#undef C
#undef P
#undef V
#undef L

}




static pthread_mutex_t logging_lock = PTHREAD_MUTEX_INITIALIZER;

RETSIGTYPE
Logging_Queues()
{
    FILE *fp;
    int i;
    struct timeval now;
    int t_nhot;

    MUTEX_LOCK(&logging_lock);

    gettimeofday(&now, NULL);

    Trace("logggin queues\n");

    fp = fopen("queues.log", "w");
    if(fp==NULL) {
        goto out;
    }

    fprintf(fp, "process %d, thread %x\n\n", getpid(), pthread_self());

        fprintf(fp, " monitor           %x\n", th_monitor);
        fprintf(fp, " listener          %x\n", th_listener);
    fprintf(fp, " manager  %d/%d\n", require_n_handler, N_HANDLER_MAX);
    for(i=0;i<require_n_handler;i++) {
        fprintf(fp, "   %3d             %x\n", i, th_handlers[i]);
    }
    fprintf(fp, "\n");

    MUTEX_LOCK(&sess_lock);
    t_nhot = nhot;
    MUTEX_UNLOCK(&sess_lock);

    fprintf(fp, "hot session %d\n\n", t_nhot);

    fprintf(fp, "== think queue\n");
    sq_show_mapfunc(fp, think_queue,    print_sess, &now);
#if 0
    fprintf(fp, "== send queue\n");
    sq_show_mapfunc(fp, send_queue,     print_sess, &now);
    fprintf(fp, "== fetch queue\n");
    sq_show_mapfunc(fp, fetch_queue,    print_sess, &now);
    fprintf(fp, "== prompt queue\n");
    sq_show_mapfunc(fp, prompt_queue,   print_sess, &now);
#endif


    report_check_mutex(fp, "mutex stat   ", &stat_lock);
    report_check_mutex(fp, "mutex sess   ", &sess_lock);
    report_check_mutex(fp, "mutex config ", &config_lock);
    report_check_mutex(fp, "mutex logging", &logging_lock);


done:

    fflush(fp);
    fclose(fp);

out:

    MUTEX_UNLOCK(&logging_lock);
}


/*
 *
 */
RETSIGTYPE
TERM_Trap()
{
    Log("; Catch signal for termination");
    MUTEX_LOCK(&stat_lock);
    catch_termination = 1;
    MUTEX_UNLOCK(&stat_lock);
    exit(127);
}

void*
signal_watcher(void *arg)
{
    int sig;
    int chk;

    while(1) {
        chk = sigwait(&handler_sig_mask, &sig);
        if(chk) {
            Trace("signal_watcher: chk %d\n", chk);
            continue;
        }

        Trace("signal_watcher: get signal %d\n", sig);

        switch(sig) {
        case SIGPIPE:
            MUTEX_LOCK(&stat_lock);
            session_queue_report_flag = 1;
            MUTEX_UNLOCK(&stat_lock);
            break;
        case SIGUSR2:
            MUTEX_LOCK(&stat_lock);
            session_queue_report_flag = 1;
            MUTEX_UNLOCK(&stat_lock);
            break;
        case SIGHUP:
            LogSync();
            break;
        case SIGINT:
        case SIGTERM:
            MUTEX_LOCK(&stat_lock);
            catch_termination = 1;
            MUTEX_UNLOCK(&stat_lock);
            Log("END");
            LogSync();
            exit(137);
            break;
        case SIGCHLD:
          {
            int stat;
            pid_t pid;

            do {
try_waitpid:
                pid = waitpid((pid_t)-1, &stat, WNOHANG | WUNTRACED);

                if(pid < 0) {
                    if(errno==EINTR) {
                        goto try_waitpid;
                    }
                    printf("error with errno=%d\n", errno);
                    break;
                }
                if (pid > 0) {
                    printf("child %d terminated\n", pid);
                }
            } while(pid>0);

          }
            break;
        default:
            break;
        }
    }

 /*NOTREACHED*/
    return (void*)NULL;
}



/*
 *
 * if fdcap is -1, this routine will set maximum value from getrlimit()
 */
void
ToBeMaximumFDCapacity(int fdcap)
{
    struct rlimit nfile;
    int last_nofile;

#ifdef RLIMIT_NOFILE
#define X   RLIMIT_NOFILE
#else
#define X   RLIMIT_OFILE
#endif

    if(getrlimit(X, &nfile)<0) {
        Error("getrlimit() for current check was fail (%d)", errno);
    }
    else {
        last_nofile = nfile.rlim_cur;
    }

    if(fdcap<0) {
        nfile.rlim_cur = nfile.rlim_max;
    }
    else {
        nfile.rlim_cur = fdcap;
    }

    if(setrlimit(X, &nfile)<0) {
        Error("settrlimit() fail (%d)", errno);
    }

    if(getrlimit(X, &nfile)<0) {
        Error("getrlimit() for verify was fail (%d)", errno);
    }
    else {
        Trace("ToBeMaximumFDCapacity: current to %d (max %d), last %d, specified %d\n",
            nfile.rlim_cur, nfile.rlim_max,
            last_nofile, fdcap);
    }

#undef X

}



static
char *compile_magic = "This program was compiled at " __DATE__ ", " __TIME__;
int listen_fd;

int
main(int argc, char **argv)
{
    int i;
    int chk;
    pthread_attr_t th_attr_high, th_attr_mid, th_attr_low;
    pthread_attr_t *p_th_attr_high, *p_th_attr_mid, *p_th_attr_low;
    struct sched_param th_param;
    int th_policy_type;

    /***
     *** Initialize
     ***/

    MUTEX_LOCK(&stat_lock);
    boot_date = time(NULL);
    MUTEX_UNLOCK(&stat_lock);

    srand(time(NULL));


    SetSignalHandler(SIGPIPE, SIG_IGN);


    /***
     *** Configuration
     ***/

    SetGlobalDefaults();
    if(getenv(DEFAULT_CONFIG_ENVNAME)!=NULL) {
        strcpy(config_filename, getenv(DEFAULT_CONFIG_ENVNAME));
    }
    if(config_filename[0]) {
        ReadConfigFile(config_filename);
    }
    ReadCommandlineOptions(argc, argv);

    FixGlobals();


    /***
     *** Logging
     ***/
    InitLogger("kotetuserv");
    Log("START %s %s %s ", KOTETU_VER,  __DATE__, __TIME__);
    LogSync();

    if(trace_flag) {
        ShowGlobals(stdout);
    }
    if(trace_flag) {
        TraceOn();
    }
    else {
        TraceOff();
    }

    date_zone_offset_init();

#ifndef NDEBUG
    Trace("not set NDEBUG, so it is debug mode\n\
If you want an actual performance, you should re-compile with -DNDEBUG\n");
#endif
#ifndef USE_3LEVEL_DIR
    Trace("not set USE_3LEVEL_DIR, so running under two level directories\n");
#endif
#ifdef USE_MLOCK
    Trace("compiled for memory locking with mlock()\n");
#endif

#if 0
    Trace("parameters:\n\
    L_URL       %6d\n\
    L_HOST      %6d\n\
    L_PATH      %6d\n\
    L_BUF       %6d\n\
    L_REQBUF    %6d\n", L_URL, L_HOST, L_PATH, L_BUF, L_REQBUF);

    Trace("size of structures:\n\
    skiplist cell   %6d\n\
    skiplist        %6d\n\
    entry           %6d\n\
    frag            %6d\n\
    bucket          %6d\n",
        sizeof(skiplist_cell_t), sizeof(skiplist_t),
        sizeof(skcache_entry_t), sizeof(skcache_frag_t),
        sizeof(bucket_t));

#endif

    ToBeMaximumFDCapacity(-1);


    /***
     *** sessions and queues
     ***/

    InitSession(require_sess_length);

#ifdef USE_WQUEUE
    think_queue_length = require_sess_length*3/4;
    sq_init(&X_think_queue,     think_queue_length);
    think_queue = &X_think_queue;
    sq_init(&X_wait_queue,     think_queue_length/4);
    think_queue = &X_wait_queue;
#else
    think_queue_length = require_sess_length;
    sq_init(&X_think_queue,     think_queue_length);
    think_queue = &X_think_queue;
#endif

#if 0
    Trace("queues:\n\
    think           %p\n\
    wait            %p\n\
    memory sweep    %p\n\
    swapout sweep   %p\n",
        think_queue, wait_queue, msweep_queue, swapout_queue);
#endif


    /***
     *** misc. statistically
     ***/

    sim_stat_alloc_new(&len_perread_stat,       "length of per read(2)",
        K(32),  server_buffer_stat_count);
    sim_stat_U_alloc_new(&object_size_stat,     "object size [KB]",
        1024, K(1),   object_size_stat_count);

    sim_stat_alloc_new(&rcode_stat,             "session response code",
        1000,   rcode_stat_count);

    sim_stat_alloc_new(&url_first_hash_stat,    "first hash value",
        K(2),   url_first_hash_stat_count);
    sim_stat_alloc_new(&url_second_hash_stat,   "second hash value",
        K(2),   url_second_hash_stat_count);
    sim_stat_alloc_new(&url_third_hash_stat,    "third hash value",
        K(2),   url_third_hash_stat_count);
    sim_stat_alloc_new(&url_fourth_hash_stat,   "fourth hash value",
        K(2),   url_fourth_hash_stat_count);

    sim_stat_alloc_new(&req_field_type_stat,    "request field type",
        256,   req_field_type_stat_count);

    sim_stat_alloc_new(&object_alloc_stat,      "object alloc",
        K(10),  object_alloc_stat_count);

    sim_stat_U_alloc_new(&ret_time,         "retrieval time [msec]", 
        50, K(5), ret_time_count);
    sim_stat_U_alloc_new(&ret_time_short,   "retrieval time (short) [msec]",
        1, K(12), ret_time_short_count);

    sim_stat_U_alloc_new(&msweep_time,            "mark time [msec]",
        1, K(10), msweep_time_count);
    sim_stat_U_alloc_new(&msweep_time_short,      "mark time (short) [msec]",
        1, K(10), msweep_time_short_count);

    /* history */
    sim_hist_alloc_new(&comp_hist,              "compaction history",
        30,  comp_hist_area);
    sim_hist_alloc_new(&expire_hist,            "expiration history",
        30,  expire_hist_area);


    /* DNS cache */
    hcache_setup();


    /***
     *** Allocate Object Cache Area
     ***/

#if 1
{
    int bk_num;
    int bk_sk_level, bk_sk_len;
    int frag_size, frag_freelist_len;
    long long U;

    bk_num = require_num_bucket;
    bk_sk_level = require_level_bucket_skiplist;
    frag_size = require_size_frag;

    skcache_calc_param(
        (long long)(require_total_memory_size-sess_memory_size),
        &bk_num, &bk_sk_level, &frag_size, (double)require_ratio_frag_length,
        &bk_sk_len, &frag_freelist_len);

    U = skcache_calc_size(&bk_num, &bk_sk_level, &bk_sk_len, &frag_freelist_len, &frag_size);

#if 0
    Trace("  => %lld B,  %ld MB\n", U,(long) dM(U));
#endif

#if 0
    Trace("@ %d %4d %2d %5d %.2f -> %9d %9d => %9d %9d\n",
        boot_date, bk_num, bk_sk_level, frag_size, (double)require_ratio_frag_length,
        bk_sk_len, frag_freelist_len,
        bk_sk_len * bk_num,  frag_freelist_len* bk_num);
#endif

#if 1
    Trace("Cache parameters:\n\
    ratio of frags              %7d\n\
    the length of frag          %7d\n\
    the number of structures\n\
      bucket                    %7d\n\
      entry                     %7d (%dx%d)\n\
      frag                      %7d\n\
\n",
        require_ratio_frag_length,
        frag_size,
        bk_num, 
        bk_sk_len * bk_num, bk_sk_len, bk_num, 
        frag_freelist_len * bk_num
);
#endif


    U = skcache_init(boot_date-86400*7,
            bk_num, bk_sk_level, bk_sk_len, frag_freelist_len, frag_size);
#if 0
    Trace("  => %lld B,  %ld MB\n", U,(long) dM(U));
#endif

    cache_memory_size = U;
}
#endif


#if 0
    Trace("Memory allocation:\n\
\ttotal        %12lld // %5ldM+%06ld\n\
\tsession      %12lld // %5ldM+%06ld %6.2f%%\n\
\tobject cache %12lld // %5ldM+%06ld %6.2f%%\n",
        (long long)require_total_memory_size,
          (long)dM(require_total_memory_size),
          (long)mM(require_total_memory_size),
        (long long)sess_memory_size,
          (long)dM(sess_memory_size), (long)mM(sess_memory_size),
          100.0*sess_memory_size/require_total_memory_size,
        (long long)cache_memory_size,
          (long)dM(cache_memory_size), (long)mM(cache_memory_size),
          100.0*cache_memory_size/require_total_memory_size);
#endif

#if 1
    mlist_show();
#endif

    nfrag_obj       = require_nfrag_obj;
    nfrag_dwrite    = require_nfrag_dwrite;
    nfrag_dread     = require_nfrag_dread;
    nfrag_nwrite    = require_nfrag_nwrite;
    nfrag_nread     = require_nfrag_nread;

#if 1
    Trace("I/O information:\n\
  Frags:\n\
    the max number of frags\n\
      (per object)\n\
      object           %3d (%7ldB %6.1fKB) // hard %3d\n\
      (per operation)\n\
      network write    %3d (%7ldB %6.1fKB) // hard %3d\n\
      network read     %3d (%7ldB %6.1fKB) // hard %3d\n\
      disk write       %3d (%7ldB %6.1fKB) // hard %3d\n\
      disk read        %3d (%7ldB %6.1fKB) // hard %3d\n",
        nfrag_obj, (long)nfrag_obj*require_size_frag,
        fdK(nfrag_obj*require_size_frag), HARD_NFRAG_OBJ,
        nfrag_nwrite, (long)nfrag_nwrite*require_size_frag, 
        fdK(nfrag_nwrite*require_size_frag), HARD_NFRAG_NWRITE,
        nfrag_nread, (long)nfrag_nread*require_size_frag, 
        fdK(nfrag_nread*require_size_frag), HARD_NFRAG_NREAD,
        nfrag_dwrite, (long)nfrag_dwrite*require_size_frag, 
        fdK(nfrag_dwrite*require_size_frag), HARD_NFRAG_DWRITE,
        nfrag_dread, (long)nfrag_dread*require_size_frag, 
        fdK(nfrag_dread*require_size_frag), HARD_NFRAG_DREAD
    );

#endif

#if 0
    {
        struct timeval now;
        gettimeofday(&now, NULL);
        client_scramble_key = now.tv_usec;
    }
#endif
#if 1
        client_scramble_key = 0xabba;
#endif


    /***
     *** Thread launching
     ***/

#ifdef DONT_USE_THREAD_ATTR
    p_th_attr_high  = NULL;
    p_th_attr_mid   = NULL;
    p_th_attr_low   = NULL;

#else
    chk = pthread_attr_init(&th_attr_high);
    chk = pthread_attr_init(&th_attr_mid);
    chk = pthread_attr_init(&th_attr_low);
    Trace("ret pthread_attr_init() %d (%d)\n", chk, errno);

    chk = pthread_attr_getschedparam(&th_attr_high, &th_param);
    Trace("default (initialized) priority value %d\n",
        th_param.sched_priority);
    th_param.sched_priority += th_param.sched_priority; /* double */
    chk = pthread_attr_setschedparam(&th_attr_high, &th_param);

    chk = pthread_attr_getschedparam(&th_attr_mid, &th_param);
    th_param.sched_priority = th_param.sched_priority; /* thru */
    chk = pthread_attr_setschedparam(&th_attr_mid, &th_param);

    chk = pthread_attr_getschedparam(&th_attr_low, &th_param);
    th_param.sched_priority = th_param.sched_priority/2; /* half */
    chk = pthread_attr_setschedparam(&th_attr_low, &th_param);

    chk = pthread_attr_getschedpolicy(&th_attr_mid, &th_policy_type);
    Trace("ret pthread_attr_getschedpolicy() %d (%d)\n", chk, errno);
    Trace("default value %d\n", th_policy_type);

#if 1
    th_policy_type = SCHED_FIFO;
    chk = pthread_attr_setschedpolicy(&th_attr_high, th_policy_type);
    th_policy_type = SCHED_RR;
    chk = pthread_attr_setschedpolicy(&th_attr_mid, th_policy_type);
    th_policy_type = SCHED_RR;
    chk = pthread_attr_setschedpolicy(&th_attr_low, th_policy_type);
#endif

    p_th_attr_high  = &th_attr_high;
    p_th_attr_mid   = &th_attr_mid;
    p_th_attr_low   = &th_attr_low;
#endif


    if(monitor_flag) {
        chk = pthread_create(&th_monitor, p_th_attr_mid,
                Monitor, NULL);
#ifdef TRACE_THREADS
        Trace("launch - monitor %d\n", chk);
#endif
    }


    /* boot listener, use maximum value of OS */
    listen_fd = TCP_bind_port(service_portno, -1);
    Trace("listen_fd %d, chk %d\n", listen_fd, chk);
    if(listen_fd<0) {
        Error("service port(%d) is already use", service_portno);
        exit(4);
    }
#ifdef DO_ACCEPT_LOCK
    chk = pthread_create(&th_listener, p_th_attr_high,
                Listener, (void*)&listen_fd);
#endif
    chk = pthread_create(&th_listener, p_th_attr_high,
                Listener, (void*)&listen_fd);


    /* boot handlers */
    for(i=0;i<require_n_handler;i++) {
        chk = pthread_create(&th_handlers[i], p_th_attr_mid,
                Slave, (void*)i);
    }

    Trace("  %d handers x %d sessions => %d concurrent sessions on time\n",
        require_n_handler, require_chunk_size,
        require_n_handler*require_chunk_size);


    /* boot disk sweeper*/
    for(i=0;i<require_n_dsweeper;i++) {
        chk = pthread_create(&th_dsweepers[i], p_th_attr_low,
                DSweeper, (void*)i);
    }



    MUTEX_LOCK(&stat_lock);
    service_start_date = time(NULL);
    MUTEX_UNLOCK(&stat_lock);


    /***
     ***
     ***/
#if 0
    pthread_sigmask(SIG_BLOCK, &handler_sig_mask, NULL);
    pthread_create(&th_signalhandler, p_th_attr_low, signal_watcher, NULL);
    pthread_sigmask(SIG_UNBLOCK, &handler_sig_mask, NULL);
#endif

    atexit(LogSync);


    MUTEX_LOCK(&stat_lock);
    Trace("initalize take %d sec(s)\n", time(NULL) - boot_date);
    MUTEX_UNLOCK(&stat_lock);


    /***
     *** forever loop
     ***/
    msleep(100);

    /* for child reaping, this thread should be awake at sometime */
    while(1) {
        struct timespec w;

        MUTEX_LOCK(&stat_lock);
        swapout_token = swapout_token_init;
        MUTEX_UNLOCK(&stat_lock);

        msleep(swapout_token_interval);

    }
}

